# SPDX-FileCopyrightText: 2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import typing as t
from pathlib import (
    Path,
)

from idf_build_apps.utils import (
    to_absolute_path,
)


class InvalidTomlError(SystemExit):
    def __init__(self, filepath: t.Union[str, Path], msg: str) -> None:
        super().__init__(f'Failed parsing toml file "{filepath}" with error: {msg}')


PYPROJECT_TOML_FN = 'pyproject.toml'
IDF_BUILD_APPS_TOML_FN = '.idf_build_apps.toml'


def load_toml(filepath: t.Union[str, Path]) -> dict:
    try:
        import tomllib  # type: ignore # python 3.11

        try:
            with open(str(filepath), 'rb') as fr:
                return tomllib.load(fr)
        except Exception as e:
            raise InvalidTomlError(filepath, str(e))
    except ImportError:
        import toml

        try:
            return toml.load(str(filepath))
        except Exception as e:
            raise InvalidTomlError(filepath, str(e))


def _get_config_from_file(filepath: str) -> t.Tuple[t.Optional[dict], str]:
    config = None
    if os.path.isfile(filepath):
        if os.path.basename(filepath) == PYPROJECT_TOML_FN:
            tool = load_toml(filepath).get('tool', None)
            if tool:
                config = tool.get('idf-build-apps', None)
        elif config is None:
            config = load_toml(filepath)

    return config, filepath


def _get_config_from_path(dirpath: str) -> t.Tuple[t.Optional[dict], str]:
    config = None
    filepath = dirpath
    if os.path.isfile(os.path.join(dirpath, PYPROJECT_TOML_FN)):
        config, filepath = _get_config_from_file(os.path.join(dirpath, PYPROJECT_TOML_FN))

    if config is None and os.path.isfile(os.path.join(dirpath, IDF_BUILD_APPS_TOML_FN)):
        config, filepath = _get_config_from_file(os.path.join(dirpath, IDF_BUILD_APPS_TOML_FN))

    return config, filepath


def get_valid_config(starts_from: str = os.getcwd(), custom_path: t.Optional[str] = None) -> t.Optional[dict]:
    root_dir = to_absolute_path('/')
    cur_dir = to_absolute_path(starts_from)

    config = None
    if custom_path and os.path.isfile(custom_path):
        config, filepath = _get_config_from_file(to_absolute_path(custom_path))
        if config is not None:
            # use print here since the verbose settings may be set in the config file
            print(f'Using custom config file: {filepath}')
            return config

    while cur_dir != root_dir and config is None:
        config, filepath = _get_config_from_path(cur_dir)
        if config is not None:
            # use print here since the verbose settings may be set in the config file
            print(f'Using config file: {filepath}')
            return config

        if os.path.exists(os.path.join(cur_dir, '.git')):
            break

        cur_dir = os.path.abspath(os.path.join(cur_dir, '..'))

    return None



# SPDX-FileCopyrightText: 2023-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

"""
The test report should look like something like this:

.. code-block:: xml

    <testsuites>
        <testsuite
                errors="0"
                failures="1"
                skipped="0"
                tests="29"
                name="Build target esp32 in example folder"
                hostname="GARM-C2-WX-1-R5S4N264"
                time="1051.215"
                timestamp="2022-12-15T13:25:17.689181">
            <testcase
                    file="examples/get_started/hello_world"
                    name="esp32.default.examples/get-started/hello_world"
                    time="60.2454">
                <failure message="build warnings or error message"></failure>
            </testcase>
            <testcase
                    file="examples/get_started/hello_world"
                    name="esp32.default.examples/get-started/hello_world"
                    time="60.2454">
                <skipped message="skipped reason"></skipped>
            </testcase>
        </testsuite>
    </testsuites>
"""

import json
import logging
import os.path
import typing as t
from datetime import (
    datetime,
    timezone,
)
from xml.etree import (
    ElementTree,
)
from xml.sax.saxutils import (
    escape,
)

from ..app import (
    App,
)
from ..constants import (
    BuildStatus,
)
from .utils import (
    get_sys_info,
)

LOGGER = logging.getLogger(__name__)


class TestCase:
    def __init__(
        self,
        name: str,
        *,
        error_reason: t.Optional[str] = None,
        failure_reason: t.Optional[str] = None,
        skipped_reason: t.Optional[str] = None,
        properties: t.Optional[t.Dict[str, str]] = None,
        duration_sec: float = 0,
        timestamp: t.Optional[datetime] = None,
    ) -> None:
        self.name = name

        self.failure_reason = failure_reason
        self.skipped_reason = skipped_reason
        self.error_reason = error_reason
        # only have one reason among these three
        if sum([self.failure_reason is not None, self.skipped_reason is not None, self.error_reason is not None]) > 1:
            raise ValueError('Only one of failure_reason, skipped_reason, error_reason can be set')

        self.duration_sec = duration_sec
        self.timestamp = timestamp or datetime.now(timezone.utc)

        self.properties = properties or {}

    @classmethod
    def from_app(cls, app: App) -> 'TestCase':
        if app.build_status in (BuildStatus.UNKNOWN, BuildStatus.SHOULD_BE_BUILT):
            raise ValueError(
                f'Cannot create build report for apps with build status {app.build_status}. '
                f'Please finish the build process first.'
            )

        kwargs: t.Dict[str, t.Any] = {
            'name': app.build_path,
            'duration_sec': app._build_duration,
            'timestamp': app._build_timestamp,
            'properties': {},
        }
        if app.build_status == BuildStatus.FAILED:
            kwargs['failure_reason'] = app.build_comment
        elif app.build_status in (BuildStatus.DISABLED, BuildStatus.SKIPPED):
            kwargs['skipped_reason'] = app.build_comment

        if app.size_json_path and os.path.isfile(app.size_json_path):
            with open(app.size_json_path) as f:
                for k, v in json.load(f).items():
                    kwargs['properties'][f'{k}'] = str(v)

        return cls(**kwargs)

    @property
    def is_failed(self) -> bool:
        return self.failure_reason is not None

    @property
    def is_skipped(self) -> bool:
        return self.skipped_reason is not None

    @property
    def is_error(self) -> bool:
        return self.error_reason is not None

    def to_xml_elem(self) -> ElementTree.Element:
        elem = ElementTree.Element(
            'testcase',
            {
                'name': self.name,
                'time': str(self.duration_sec),
                'timestamp': self.timestamp.isoformat(),
            },
        )
        if self.error_reason:
            ElementTree.SubElement(elem, 'error', {'message': escape(self.error_reason)})
        elif self.failure_reason:
            ElementTree.SubElement(elem, 'failure', {'message': escape(self.failure_reason)})
        elif self.skipped_reason:
            ElementTree.SubElement(elem, 'skipped', {'message': escape(self.skipped_reason)})

        if self.properties:
            for k, v in self.properties.items():
                elem.attrib[k] = escape(str(v))

        return elem


class TestSuite:
    def __init__(self, name: str) -> None:
        self.name = name

        self.test_cases: t.List[TestCase] = []

        self.tests = 0  # passed, actually
        self.errors = 0  # setup error
        self.failures = 0  # runtime failures
        self.skipped = 0

        self.duration_sec: float = 0
        self.timestamp = datetime.now(timezone.utc)

        self.properties = get_sys_info()

    def add_test_case(self, test_case: TestCase) -> None:
        self.test_cases.append(test_case)

        if test_case.is_error:
            self.errors += 1
        elif test_case.is_failed:
            self.failures += 1
        elif test_case.is_skipped:
            self.skipped += 1
        else:
            self.tests += 1

        self.duration_sec += test_case.duration_sec

    def to_xml_elem(self) -> ElementTree.Element:
        elem = ElementTree.Element(
            'testsuite',
            {
                'name': self.name,
                'tests': str(self.tests),
                'errors': str(self.errors),
                'failures': str(self.failures),
                'skipped': str(self.skipped),
                'time': str(self.duration_sec),
                'timestamp': self.timestamp.isoformat(),
                **self.properties,
            },
        )

        for test_case in self.test_cases:
            elem.append(test_case.to_xml_elem())

        return elem


class TestReport:
    def __init__(self, test_suites: t.List[TestSuite], filepath: str) -> None:
        self.test_suites: t.List[TestSuite] = test_suites

        self.filepath = filepath

    def create_test_report(self) -> None:
        xml = ElementTree.Element('testsuites')

        for test_suite in self.test_suites:
            xml.append(test_suite.to_xml_elem())

        ElementTree.ElementTree(xml).write(self.filepath, encoding='utf-8')
        LOGGER.info('Test report written to %s', self.filepath)



# SPDX-FileCopyrightText: 2023-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import shutil
import subprocess

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'idf-build-apps'
project_homepage = 'https://github.com/espressif/idf-build-apps'
copyright = '2023, Espressif Systems (Shanghai) Co., Ltd.'  # noqa: A001
author = 'Fu Hanxi'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx_copybutton',
    'myst_parser',
    'sphinxcontrib.mermaid',
    'sphinxarg.ext',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_css_files = ['theme_overrides.css']
html_logo = '_static/espressif-logo.svg'
html_static_path = ['_static']
html_theme = 'sphinx_rtd_theme'

# mermaid 10.2.0 will show syntax error
# use fixed version instead
mermaid_version = '10.6.1'

# building docs
os.environ['BUILDING_DOCS'] = '1'

# generating api docs
docs_dir = os.path.dirname(__file__)
api_dir = os.path.join(docs_dir, 'api')
if os.path.isdir(api_dir):
    shutil.rmtree(api_dir)
subprocess.run(
    [
        'sphinx-apidoc',
        os.path.join(docs_dir, '..', 'idf_build_apps'),
        '-f',
        '-H',
        'API Reference',
        '--no-headings',
        '-t',
        '_apidoc_templates',
        '-o',
        api_dir,
    ]
)



# SPDX-FileCopyrightText: 2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import typing as t

import yaml


def parse_postfixes(manifest_dict: t.Dict):
    for folder, folder_rule in manifest_dict.items():
        if folder.startswith('.'):
            continue

        if not folder_rule:
            continue

        updated_folder: t.Dict = {}
        sorted_keys = sorted(folder_rule)
        for key in sorted_keys:
            if not key.endswith(('+', '-')):
                updated_folder[key] = folder_rule[key]
                continue

            operation = key[-1]

            if_dict_obj = []
            other_dict_obj = []
            str_obj = set()
            for obj in updated_folder[key[:-1]]:
                if isinstance(obj, t.Dict):
                    if 'if' in obj:
                        if_dict_obj.append(obj)
                    else:
                        other_dict_obj.append(obj)
                else:
                    str_obj.add(obj)

            for obj in folder_rule[key]:
                if isinstance(obj, t.Dict):
                    _l = obj['if']
                    if isinstance(_l, str):
                        _l = _l.replace(' ', '')

                    new_dict_obj = []
                    for obj_j in if_dict_obj:
                        _r = obj_j['if']
                        if isinstance(_r, str):
                            _r = _r.replace(' ', '')
                        if _l != _r:
                            new_dict_obj.append(obj_j)
                    if_dict_obj = new_dict_obj

                    if operation == '+':
                        if_dict_obj.append(obj)
                else:
                    str_obj.add(obj) if operation == '+' else str_obj.remove(obj)

            updated_folder[key[:-1]] = if_dict_obj + other_dict_obj + sorted(str_obj)

        manifest_dict[folder] = updated_folder


def parse(path: str) -> t.Dict:
    with open(path) as f:
        manifest_dict = yaml.safe_load(f) or {}
    parse_postfixes(manifest_dict)
    return manifest_dict



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import os

import pytest

import idf_build_apps
from idf_build_apps import (
    App,
    setup_logging,
)


@pytest.fixture(autouse=True)
def clean_cls_attr():
    App.MANIFEST = None
    idf_build_apps.SESSION_ARGS.clean()


@pytest.fixture(autouse=True)
def setup_logging_debug():
    setup_logging(1)


def create_project(name, folder):
    p = str(folder / name)
    os.makedirs(p)
    os.makedirs(os.path.join(p, 'main'))

    with open(os.path.join(p, 'CMakeLists.txt'), 'w') as fw:
        fw.write(
            f"""cmake_minimum_required(VERSION 3.16)
include($ENV{{IDF_PATH}}/tools/cmake/project.cmake)
project({name})
"""
        )

    with open(os.path.join(p, 'main', 'CMakeLists.txt'), 'w') as fw:
        fw.write(
            f"""idf_component_register(SRCS "{name}.c"
INCLUDE_DIRS ".")
"""
        )

    with open(os.path.join(p, 'main', f'{name}.c'), 'w') as fw:
        fw.write(
            """#include <stdio.h>
void app_main(void) {}
"""
        )



# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import os.path
import re
import typing as t
from pathlib import (
    Path,
)

from .app import (
    App,
    CMakeApp,
)
from .constants import (
    BuildStatus,
)
from .utils import (
    config_rules_from_str,
    to_absolute_path,
    to_list,
)

LOGGER = logging.getLogger(__name__)


def _get_apps_from_path(
    path: str,
    target: str,
    app_cls: t.Type[App] = CMakeApp,
    work_dir: t.Optional[str] = None,
    build_dir: str = 'build',
    config_rules_str: t.Optional[t.List[str]] = None,
    build_log_filename: t.Optional[str] = None,
    size_json_filename: t.Optional[str] = None,
    check_warnings: bool = False,
    preserve: bool = True,
    manifest_rootpath: t.Optional[str] = None,
    modified_components: t.Optional[t.List[str]] = None,
    modified_files: t.Optional[t.List[str]] = None,
    check_app_dependencies: bool = False,
    sdkconfig_defaults_str: t.Optional[str] = None,
    include_skipped_apps: bool = False,
    include_disabled_apps: bool = False,
) -> t.List[App]:
    modified_components = to_list(modified_components)
    modified_files = to_list(modified_files)

    def _validate_app(_app: App) -> bool:
        if target not in _app.supported_targets:
            LOGGER.debug('=> Ignored. %s only supports targets: %s', _app, ', '.join(_app.supported_targets))
            _app.build_status = BuildStatus.DISABLED
            return include_disabled_apps

        _app._check_should_build(
            manifest_rootpath=manifest_rootpath,
            modified_components=modified_components,
            modified_files=modified_files,
            check_app_dependencies=check_app_dependencies,
        )

        # for unknown ones, we keep them to the build stage to judge
        if _app.build_status == BuildStatus.SKIPPED:
            return include_skipped_apps

        return True

    if not app_cls.is_app(path):
        LOGGER.debug('Skipping. %s is not an app', path)
        return []

    config_rules = config_rules_from_str(config_rules_str)
    if not config_rules:
        config_rules = []

    apps = []
    default_config_name = ''
    sdkconfig_paths_matched = False
    for rule in config_rules:
        if not rule.file_name:
            default_config_name = rule.config_name
            continue

        sdkconfig_paths = sorted([str(p.resolve()) for p in Path(path).glob(rule.file_name)])

        if sdkconfig_paths:
            sdkconfig_paths_matched = True  # skip the next block for no wildcard config rules

        for sdkconfig_path in sdkconfig_paths:
            if sdkconfig_path.endswith(f'.{target}'):
                LOGGER.debug('=> Skipping sdkconfig %s which is target-specific', sdkconfig_path)
                continue

            # Figure out the config name
            config_name = rule.config_name or ''
            if '*' in rule.file_name:
                # convert glob pattern into a regex
                regex_str = r'.*' + rule.file_name.replace('.', r'\.').replace('*', r'(.*)')
                groups = re.match(regex_str, sdkconfig_path)
                assert groups
                config_name = groups.group(1)

            app = app_cls(
                path,
                target,
                sdkconfig_path=sdkconfig_path,
                config_name=config_name,
                work_dir=work_dir,
                build_dir=build_dir,
                build_log_filename=build_log_filename,
                size_json_filename=size_json_filename,
                check_warnings=check_warnings,
                preserve=preserve,
                sdkconfig_defaults_str=sdkconfig_defaults_str,
            )
            if _validate_app(app):
                LOGGER.debug('Found app: %s', app)
                apps.append(app)

            LOGGER.debug('')  # add one empty line for separating different finds

    # no config rules matched, use default app
    if not sdkconfig_paths_matched:
        app = app_cls(
            path,
            target,
            sdkconfig_path=None,
            config_name=default_config_name,
            work_dir=work_dir,
            build_dir=build_dir,
            build_log_filename=build_log_filename,
            size_json_filename=size_json_filename,
            check_warnings=check_warnings,
            preserve=preserve,
            sdkconfig_defaults_str=sdkconfig_defaults_str,
        )

        if _validate_app(app):
            LOGGER.debug('Found app: %s', app)
            apps.append(app)

        LOGGER.debug('')  # add one empty line for separating different finds

    return sorted(apps)


def _find_apps(
    path: str,
    target: str,
    app_cls: t.Type[App] = CMakeApp,
    recursive: bool = False,
    exclude_list: t.Optional[t.List[str]] = None,
    **kwargs,
) -> t.List[App]:
    exclude_list = exclude_list or []
    LOGGER.debug(
        'Looking for %s apps in %s%s with target %s',
        app_cls.__name__,
        path,
        ' recursively' if recursive else '',
        target,
    )

    if not recursive:
        if exclude_list:
            LOGGER.warning('--exclude option is ignored when used without --recursive')

        return _get_apps_from_path(path, target, app_cls, **kwargs)

    # The remaining part is for recursive == True
    apps = []
    # handle the exclude list, since the config file might use linux style, but run in windows
    exclude_paths_list = [to_absolute_path(p) for p in exclude_list]
    for root, dirs, _ in os.walk(path):
        LOGGER.debug('Entering %s', root)
        root_path = to_absolute_path(root)
        if root_path in exclude_paths_list:
            LOGGER.debug('=> Skipping %s (excluded)', root)
            del dirs[:]
            continue

        if os.path.basename(root_path) == 'managed_components':  # idf-component-manager
            LOGGER.debug('=> Skipping %s (managed components)', root_path)
            del dirs[:]
            continue

        _found_apps = _get_apps_from_path(root, target, app_cls, **kwargs)
        if _found_apps:  # root has at least one app
            LOGGER.debug('=> Stop iteration sub dirs of %s since it has apps', root)
            del dirs[:]
            apps.extend(_found_apps)
            continue

    return apps



# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import os
from pathlib import (
    Path,
)

import pytest

from idf_build_apps.utils import (
    files_matches_patterns,
    get_parallel_start_stop,
    rmdir,
    to_absolute_path,
)


@pytest.mark.parametrize(
    'patterns, expected',
    [
        ('*.txt', ['test/inner', 'test/inner/test.txt', 'test/test.txt']),
        (
            ['*.txt', '*.log'],
            [
                'test/inner',
                'test/inner/build.log',
                'test/inner/test.txt',
                'test/test.txt',
            ],
        ),
    ],
)
def test_rmdir(tmpdir, patterns, expected):
    test_dir = tmpdir.mkdir('test')
    dir1 = test_dir.mkdir('inner')
    test_dir.mkdir('inner2')

    Path(dir1 / 'test.txt').touch()
    Path(dir1 / 'build.log').touch()
    Path(test_dir / 'test.txt').touch()

    rmdir(test_dir, exclude_file_patterns=patterns)

    assert sorted(Path(test_dir).glob('**/*')) == [Path(tmpdir / i) for i in expected]


@pytest.mark.parametrize(
    'total, parallel_count, parallel_index, start, stop',
    [
        (1, 1, 1, 1, 1),
        (1, 2, 2, 2, 1),
        (1, 10, 1, 1, 1),
        (6, 4, 1, 1, 2),
        (6, 4, 2, 3, 4),
        (6, 4, 3, 5, 6),
        (6, 4, 4, 7, 6),
        (10, 10, 9, 9, 9),
        (33, 2, 1, 1, 17),
        (33, 2, 2, 18, 33),
    ],
)
def test_get_parallel_start_stop(total, parallel_count, parallel_index, start, stop):
    assert (start, stop) == get_parallel_start_stop(total, parallel_count, parallel_index)


def test_files_matches_patterns(tmpdir):
    # used for testing absolute paths
    temp_dir = tmpdir.mkdir('temp')
    os.chdir(temp_dir)

    # create real files
    test_dir = tmpdir.mkdir('test')
    a_dir = test_dir.mkdir('a')
    b_dir = a_dir.mkdir('b')
    c_dir = b_dir.mkdir('c')
    b_py = Path(a_dir / 'b.py')
    c_py = Path(b_dir / 'c.py')
    d_py = Path(c_dir / 'd.py')
    b_py.touch()
    c_py.touch()
    d_py.touch()

    # ├── temp
    # └── test
    #     └── a
    #         ├── b
    #         │   ├── c
    #         │   │   └── d.py
    #         │   └── c.py
    #         ├── .hidden
    #         └── b.py
    #

    # in correct relative path
    for matched_files, pat, rootpath in [
        ([b_py], '*.py', a_dir),
        ([b_py, c_py, d_py], '**/*.py', a_dir),
        ([c_py], '*.py', b_dir),
        ([c_py, d_py], '**/*.py', b_dir),
        ([d_py], '*.py', c_dir),
        ([d_py], '**/*.py', c_dir),
    ]:
        for f in matched_files:
            assert files_matches_patterns(f, pat, rootpath)

    # in None root path with relative pattern
    for matched_files, pat in [
        ([b_py], 'a/*.py'),
        ([b_py, c_py, d_py], 'a/**/*.py'),
        ([c_py], 'a/b/*.py'),
        ([c_py, d_py], 'a/b/**/*.py'),
        ([d_py], 'a/b/c/*.py'),
        ([d_py], 'a/b/c/**/*.py'),
    ]:
        for f in matched_files:
            # with correct pwd
            os.chdir(test_dir)
            assert files_matches_patterns(f, pat)

            # with wrong pwd
            os.chdir(temp_dir)
            assert not files_matches_patterns(f, pat)

    # use correct absolute path, in wrong pwd
    for matched_files, pat in [
        ([b_py], 'a/*.py'),
        ([b_py, c_py, d_py], 'a/**/*.py'),
        ([c_py], 'a/b/*.py'),
        ([c_py, d_py], 'a/b/**/*.py'),
        ([d_py], 'a/b/c/*.py'),
        ([d_py], 'a/b/c/**/*.py'),
    ]:
        abs_pat = to_absolute_path(pat, test_dir)
        os.chdir(temp_dir)
        for f in matched_files:
            assert files_matches_patterns(f, abs_pat)



# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

"""
Tools for building ESP-IDF related apps.
"""

# ruff: noqa: E402
# avoid circular imports

__version__ = '2.4.2'

from .session_args import (
    SessionArgs,
)

SESSION_ARGS = SessionArgs()

from .app import (
    App,
    AppDeserializer,
    CMakeApp,
    MakeApp,
)
from .log import (
    setup_logging,
)
from .main import (
    build_apps,
    find_apps,
    json_to_app,
)

__all__ = [
    'App',
    'AppDeserializer',
    'CMakeApp',
    'MakeApp',
    'build_apps',
    'find_apps',
    'json_to_app',
    'setup_logging',
]



# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import enum
import importlib
import logging
import os
import re
import sys
import tempfile
import typing as t

from .utils import (
    to_version,
)

LOGGER = logging.getLogger(__name__)

_BUILDING_DOCS = bool(os.getenv('BUILDING_DOCS'))
if _BUILDING_DOCS:
    print('Building Docs... Faking lots of constant values')


if _BUILDING_DOCS:
    _idf_env = tempfile.gettempdir()
else:
    _idf_env = os.getenv('IDF_PATH') or ''
    if not _idf_env:
        raise SystemExit('environment variable IDF_PATH must be set')


IDF_PATH = os.path.abspath(_idf_env)
IDF_PY = os.path.join(IDF_PATH, 'tools', 'idf.py')
IDF_SIZE_PY = os.path.join(IDF_PATH, 'tools', 'idf_size.py')
PROJECT_DESCRIPTION_JSON = 'project_description.json'
DEFAULT_SDKCONFIG = 'sdkconfig.defaults'


sys.path.append(os.path.join(IDF_PATH, 'tools', 'idf_py_actions'))
if _BUILDING_DOCS:
    _idf_py_constant_py = object()
else:
    try:
        _idf_py_constant_py = importlib.import_module('constants')
    except ModuleNotFoundError:
        LOGGER.warning('Cannot import constants from idf_py_actions')
        _idf_py_constant_py = object()
SUPPORTED_TARGETS = getattr(_idf_py_constant_py, 'SUPPORTED_TARGETS', [])
PREVIEW_TARGETS = getattr(_idf_py_constant_py, 'PREVIEW_TARGETS', [])
ALL_TARGETS = SUPPORTED_TARGETS + PREVIEW_TARGETS


def _idf_version_from_cmake() -> t.Tuple[int, int, int]:
    version_path = os.path.join(IDF_PATH, 'tools', 'cmake', 'version.cmake')
    if not os.path.isfile(version_path):
        raise ValueError(f'File {version_path} does not exist')

    regex = re.compile(r'^\s*set\s*\(\s*IDF_VERSION_([A-Z]{5})\s+(\d+)')
    ver = {}
    try:
        with open(version_path) as f:
            for line in f:
                m = regex.match(line)

                if m:
                    ver[m.group(1)] = m.group(2)

        return int(ver['MAJOR']), int(ver['MINOR']), int(ver['PATCH'])
    except (KeyError, OSError):
        raise ValueError(f'Cannot find ESP-IDF version in {version_path}')


if _BUILDING_DOCS:
    IDF_VERSION_MAJOR, IDF_VERSION_MINOR, IDF_VERSION_PATCH = 1, 0, 0
else:
    IDF_VERSION_MAJOR, IDF_VERSION_MINOR, IDF_VERSION_PATCH = _idf_version_from_cmake()

IDF_VERSION = to_version(f'{IDF_VERSION_MAJOR}.{IDF_VERSION_MINOR}.{IDF_VERSION_PATCH}')


class BuildStatus(str, enum.Enum):
    UNKNOWN = 'unknown'
    DISABLED = 'disabled'
    SKIPPED = 'skipped'
    SHOULD_BE_BUILT = 'should be built'
    FAILED = 'build failed'
    SUCCESS = 'build success'


class BuildStage(str, enum.Enum):
    DRY_RUN = 'Dry Run'
    PRE_BUILD = 'Pre Build'
    BUILD = 'Build'
    POST_BUILD = 'Post Build'

    @classmethod
    def max_length(cls) -> int:
        return max(len(v.value) for v in cls.__members__.values())


completion_instructions = """
With the `--activate` option, detect your shell type and add the appropriate commands to your shell's config file
so that it runs on startup. You will likely have to restart.
or re-login for the autocompletion to start working.

You can also specify your shell using the `--shell` option.

If you do not want automatic modification of your shell configuration file
You can manually add the commands provided below to activate autocompletion.
or run them in your current terminal session for one-time activation.

Once again, you will likely have to restart
or re-login for the autocompletion to start working.

bash:
    eval "$(register-python-argcomplete idf-build-apps)"

zsh:
    To activate completions in zsh, first make sure compinit is marked for
    autoload and run autoload:

    autoload -U compinit
    compinit

    Afterwards you can enable completions for idf-build-apps:

    eval "$(register-python-argcomplete idf-build-apps)"

fish:
    # Not required to be in the config file, only run once
    register-python-argcomplete --shell fish idf-build-apps >~/.config/fish/completions/idf-build-apps.fish
"""



# PYTHON_ARGCOMPLETE_OK

# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import argparse
import json
import logging
import os
import re
import sys
import textwrap
import typing as t
from copy import deepcopy

import argcomplete
from pydantic import (
    Field,
    create_model,
)

from . import (
    SESSION_ARGS,
)
from .app import (
    App,
    AppDeserializer,
    CMakeApp,
    MakeApp,
)
from .autocompletions import activate_completions
from .build_apps_args import (
    BuildAppsArgs,
)
from .config import (
    get_valid_config,
)
from .constants import ALL_TARGETS, BuildStatus, completion_instructions
from .finder import (
    _find_apps,
)
from .junit import (
    TestCase,
    TestReport,
    TestSuite,
)
from .log import (
    setup_logging,
)
from .manifest.manifest import (
    FolderRule,
    Manifest,
)
from .utils import (
    AutocompleteActivationError,
    InvalidCommand,
    files_matches_patterns,
    get_parallel_start_stop,
    semicolon_separated_str_to_list,
    to_absolute_path,
    to_list,
)

LOGGER = logging.getLogger(__name__)


def _check_app_dependency(
    manifest_rootpath: t.Optional[str] = None,
    modified_components: t.Optional[t.List[str]] = None,
    modified_files: t.Optional[t.List[str]] = None,
    ignore_app_dependencies_components: t.Optional[t.List[str]] = None,
    ignore_app_dependencies_filepatterns: t.Optional[t.List[str]] = None,
) -> bool:
    # not check since modified_components and modified_files are not passed
    if modified_components is None and modified_files is None:
        return False

    # not check since ignore_app_dependencies_components is passed and matched
    if (
        ignore_app_dependencies_components
        and modified_components is not None
        and set(modified_components).intersection(ignore_app_dependencies_components)
    ):
        LOGGER.info(
            'Build all apps since modified components %s matches ignored components %s',
            ', '.join(modified_components),
            ', '.join(ignore_app_dependencies_components),
        )
        return False

    # not check since ignore_app_dependencies_filepatterns is passed and matched
    if (
        ignore_app_dependencies_filepatterns
        and modified_files is not None
        and files_matches_patterns(modified_files, ignore_app_dependencies_filepatterns, manifest_rootpath)
    ):
        LOGGER.info(
            'Build all apps since modified files %s matches ignored file patterns %s',
            ', '.join(modified_files),
            ', '.join(ignore_app_dependencies_filepatterns),
        )
        return False

    return True


def find_apps(
    paths: t.Union[t.List[str], str],
    target: str,
    *,
    build_system: t.Union[t.Type[App], str] = CMakeApp,
    recursive: bool = False,
    exclude_list: t.Optional[t.List[str]] = None,
    work_dir: t.Optional[str] = None,
    build_dir: str = 'build',
    config_rules_str: t.Optional[t.Union[t.List[str], str]] = None,
    build_log_filename: t.Optional[str] = None,
    size_json_filename: t.Optional[str] = None,
    check_warnings: bool = False,
    preserve: bool = True,
    manifest_rootpath: t.Optional[str] = None,
    manifest_files: t.Optional[t.Union[t.List[str], str]] = None,
    check_manifest_rules: bool = False,
    default_build_targets: t.Optional[t.Union[t.List[str], str]] = None,
    modified_components: t.Optional[t.Union[t.List[str], str]] = None,
    modified_files: t.Optional[t.Union[t.List[str], str]] = None,
    ignore_app_dependencies_components: t.Optional[t.Union[t.List[str], str]] = None,
    ignore_app_dependencies_filepatterns: t.Optional[t.Union[t.List[str], str]] = None,
    sdkconfig_defaults: t.Optional[str] = None,
    include_skipped_apps: bool = False,
    include_disabled_apps: bool = False,
) -> t.List[App]:
    """
    Find app directories in paths (possibly recursively), which contain apps for the given build system, compatible
    with the given target

    :param paths: list of app directories (can be / usually will be a relative path)
    :param target: desired value of IDF_TARGET; apps incompatible with the given target are skipped.
    :param build_system: class of the build system, default CMakeApp
    :param recursive: Recursively search into the nested sub-folders if no app is found or not
    :param exclude_list: list of paths to be excluded from the recursive search
    :param work_dir: directory where the app should be copied before building. Support placeholders
    :param build_dir: directory where the build will be done. Support placeholders.
    :param config_rules_str: mapping of sdkconfig file name patterns to configuration names
    :param build_log_filename: filename of the build log. Will be placed under the app.build_path.
        Support placeholders. The logs will go to stdout/stderr if not specified
    :param size_json_filename: filename to collect the app's size information. Will be placed under the app.build_path.
        Support placeholders. The app's size information won't be collected if not specified
    :param check_warnings: Check for warnings in the build log or not
    :param preserve: Preserve the built binaries or not
    :param manifest_rootpath: The root path of the manifest files. Usually the folders specified in the manifest files
        are relative paths. Use the current directory if not specified
    :param manifest_files: paths of the manifest files
    :param check_manifest_rules: check the manifest rules or not
    :param default_build_targets: default build targets used in manifest files
    :param modified_components: modified components
    :param modified_files: modified files
    :param ignore_app_dependencies_components: components used for ignoring checking the app dependencies
    :param ignore_app_dependencies_filepatterns: file patterns used for ignoring checking the app dependencies
    :param sdkconfig_defaults: semicolon-separated string, pass to idf.py -DSDKCONFIG_DEFAULTS if specified,
        also could be set via environment variables "SDKCONFIG_DEFAULTS"
    :param include_skipped_apps: include skipped apps or not
    :param include_disabled_apps: include disabled apps or not
    :return: list of found apps
    """
    if default_build_targets:
        default_build_targets = to_list(default_build_targets)
        LOGGER.info('Overriding default build targets to %s', default_build_targets)
        FolderRule.DEFAULT_BUILD_TARGETS = default_build_targets

    if isinstance(build_system, str):
        # backwards compatible
        if build_system == 'cmake':
            build_system = CMakeApp
        elif build_system == 'make':
            build_system = MakeApp
        else:
            raise ValueError('Only Support "make" and "cmake"')
    app_cls = build_system

    # always set the manifest rootpath at the very beginning of find_apps in case ESP-IDF switches the branch.
    Manifest.ROOTPATH = to_absolute_path(manifest_rootpath or os.curdir)
    Manifest.CHECK_MANIFEST_RULES = check_manifest_rules

    if manifest_files:
        App.MANIFEST = Manifest.from_files(to_list(manifest_files))

    modified_components = to_list(modified_components)
    modified_files = to_list(modified_files)
    ignore_app_dependencies_components = to_list(ignore_app_dependencies_components)
    ignore_app_dependencies_filepatterns = to_list(ignore_app_dependencies_filepatterns)
    config_rules_str = to_list(config_rules_str)

    apps = []
    if target == 'all':
        targets = ALL_TARGETS
    else:
        targets = [target]

    for target in targets:
        for path in to_list(paths):
            path = path.strip()
            apps.extend(
                _find_apps(
                    path,
                    target,
                    app_cls,
                    recursive,
                    exclude_list or [],
                    work_dir=work_dir,
                    build_dir=build_dir or 'build',
                    config_rules_str=config_rules_str,
                    build_log_filename=build_log_filename,
                    size_json_filename=size_json_filename,
                    check_warnings=check_warnings,
                    preserve=preserve,
                    manifest_rootpath=manifest_rootpath,
                    check_app_dependencies=_check_app_dependency(
                        manifest_rootpath=manifest_rootpath,
                        modified_components=modified_components,
                        modified_files=modified_files,
                        ignore_app_dependencies_components=ignore_app_dependencies_components,
                        ignore_app_dependencies_filepatterns=ignore_app_dependencies_filepatterns,
                    ),
                    modified_components=modified_components,
                    modified_files=modified_files,
                    sdkconfig_defaults_str=sdkconfig_defaults,
                    include_skipped_apps=include_skipped_apps,
                    include_disabled_apps=include_disabled_apps,
                )
            )

    LOGGER.info(f'Found {len(apps)} apps in total')

    return sorted(apps)


def build_apps(
    apps: t.Union[t.List[App], App],
    *,
    build_verbose: bool = False,
    dry_run: bool = False,
    keep_going: bool = False,
    ignore_warning_strs: t.Optional[t.List[str]] = None,
    ignore_warning_file: t.Optional[t.TextIO] = None,
    copy_sdkconfig: bool = False,
    manifest_rootpath: t.Optional[str] = None,
    modified_components: t.Optional[t.Union[t.List[str], str]] = None,
    modified_files: t.Optional[t.Union[t.List[str], str]] = None,
    ignore_app_dependencies_components: t.Optional[t.Union[t.List[str], str]] = None,
    ignore_app_dependencies_filepatterns: t.Optional[t.Union[t.List[str], str]] = None,
    check_app_dependencies: t.Optional[bool] = None,
    # BuildAppsArgs
    parallel_count: int = 1,
    parallel_index: int = 1,
    collect_size_info: t.Optional[str] = None,
    collect_app_info: t.Optional[str] = None,
    junitxml: t.Optional[str] = None,
) -> int:
    """
    Build all the specified apps

    :param apps: list of apps to be built
    :param build_verbose: call ``--verbose`` in ``idf.py build`` or not
    :param dry_run: simulate this run or not
    :param keep_going: keep building or not if one app's build failed
    :param ignore_warning_strs: ignore build warnings that matches any of the specified regex patterns
    :param ignore_warning_file: ignore build warnings that matches any of the lines of the regex patterns in the
        specified file
    :param copy_sdkconfig: copy the sdkconfig file to the build directory or not
    :param manifest_rootpath: The root path of the manifest files. Usually the folders specified in the manifest files
        are relative paths. Use the current directory if not specified
    :param modified_components: modified components
    :param modified_files: modified files
    :param ignore_app_dependencies_components: components used for ignoring checking the app dependencies
    :param ignore_app_dependencies_filepatterns: file patterns used for ignoring checking the app dependencies
    :param check_app_dependencies: check app dependencies or not. if not set, will be calculated by modified_components,
        modified_files, and ignore_app_dependencies_filepatterns
    :param parallel_count: number of parallel tasks to run
    :param parallel_index: index of the parallel task to run
    :param collect_size_info: file path to record all generated size files' paths if specified
    :param collect_app_info: file path to record all the built apps' info if specified
    :param junitxml: path of the junitxml file
    :return: exit code
    """
    apps = to_list(apps)
    modified_components = to_list(modified_components)
    modified_files = to_list(modified_files)
    ignore_app_dependencies_components = to_list(ignore_app_dependencies_components)
    ignore_app_dependencies_filepatterns = to_list(ignore_app_dependencies_filepatterns)

    test_suite = TestSuite('build_apps')

    ignore_warnings_regexes = []
    if ignore_warning_strs:
        for s in ignore_warning_strs:
            ignore_warnings_regexes.append(re.compile(s))
    if ignore_warning_file:
        for s in ignore_warning_file:
            ignore_warnings_regexes.append(re.compile(s.strip()))
    App.IGNORE_WARNS_REGEXES = ignore_warnings_regexes

    start, stop = get_parallel_start_stop(len(apps), parallel_count, parallel_index)
    LOGGER.info('Total %s apps. running build for app %s-%s', len(apps), start, stop)

    build_apps_args = BuildAppsArgs(
        parallel_count=parallel_count,
        parallel_index=parallel_index,
        collect_size_info=collect_size_info,
        collect_app_info=collect_app_info,
        junitxml=junitxml,
    )
    for app in apps[start - 1 : stop]:  # we use 1-based
        app.build_apps_args = build_apps_args

    # cleanup collect files if exists at this early-stage
    for f in (build_apps_args.collect_app_info, build_apps_args.collect_size_info, build_apps_args.junitxml):
        if f and os.path.isfile(f):
            os.remove(f)
            LOGGER.debug('Remove existing collect file %s', f)

    exit_code = 0
    for i, app in enumerate(apps):
        index = i + 1  # we use 1-based
        if index < start or index > stop:
            continue

        # attrs
        app.dry_run = dry_run
        app.index = index
        app.verbose = build_verbose
        app.copy_sdkconfig = copy_sdkconfig

        LOGGER.info('(%s/%s) Building app: %s', index, len(apps), app)

        app.build(
            manifest_rootpath=manifest_rootpath,
            modified_components=modified_components,
            modified_files=modified_files,
            check_app_dependencies=_check_app_dependency(
                manifest_rootpath=manifest_rootpath,
                modified_components=modified_components,
                modified_files=modified_files,
                ignore_app_dependencies_components=ignore_app_dependencies_components,
                ignore_app_dependencies_filepatterns=ignore_app_dependencies_filepatterns,
            )
            if check_app_dependencies is None
            else check_app_dependencies,
        )
        test_suite.add_test_case(TestCase.from_app(app))

        if app.build_comment:
            LOGGER.info('%s (%s)', app.build_status.value, app.build_comment)
        else:
            LOGGER.info('%s', app.build_status.value)

        if build_apps_args.collect_app_info:
            with open(build_apps_args.collect_app_info, 'a') as fw:
                fw.write(app.to_json() + '\n')
            LOGGER.debug('Recorded app info in %s', build_apps_args.collect_app_info)

        if app.build_status == BuildStatus.FAILED:
            if not keep_going:
                return 1
            else:
                exit_code = 1
        elif app.build_status == BuildStatus.SUCCESS:
            if build_apps_args.collect_size_info and app.size_json_path:
                if os.path.isfile(app.size_json_path):
                    with open(build_apps_args.collect_size_info, 'a') as fw:
                        fw.write(
                            json.dumps(
                                {
                                    'app_name': app.name,
                                    'config_name': app.config_name,
                                    'target': app.target,
                                    'path': app.size_json_path,
                                }
                            )
                            + '\n'
                        )
                    LOGGER.debug('Recorded size info file path in %s', build_apps_args.collect_size_info)

        LOGGER.info('')  # add one empty line for separating different builds

    if build_apps_args.junitxml:
        TestReport([test_suite], build_apps_args.junitxml).create_test_report()
        LOGGER.info('Generated junit report for build apps: %s', build_apps_args.junitxml)

    return exit_code


class IdfBuildAppsCliFormatter(argparse.HelpFormatter):
    LINE_SEP = '$LINE_SEP$'

    def _split_lines(self, text, width):
        parts = text.split(self.LINE_SEP)

        text = self._whitespace_matcher.sub(' ', parts[0]).strip()
        return textwrap.wrap(text, width) + parts[1:]

    def _get_help_string(self, action):
        """
        Add the default value to the option help message.

        ArgumentDefaultsHelpFormatter and BooleanOptionalAction when it isn't
        already present. This code will do that, detecting corner cases to
        prevent duplicates or cases where it wouldn't make sense to the end
        user.
        """
        _help = action.help
        if _help is None:
            _help = ''

        if action.dest == 'config_file':
            return _help

        if action.default is not argparse.SUPPRESS:
            if action.default is None:
                default_type = str
            else:
                default_type = type(action.default)

            if action.nargs in [argparse.ZERO_OR_MORE, argparse.ONE_OR_MORE]:
                _type = f'list[{default_type.__name__}]'
            else:
                _type = default_type.__name__

            defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
            if action.option_strings or action.nargs in defaulting_nargs:
                _help += f'{self.LINE_SEP} - default: %(default)s'

            _help += f'{self.LINE_SEP} - config name: {action.dest}'
            _help += f'{self.LINE_SEP} - config type: {_type}'

        return _help


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description='Tools for building ESP-IDF related apps.'
        'Some CLI options can be expanded by the following placeholders, like "--work-dir", "--build-dir", etc.:\n'
        '- @t: would be replaced by the target chip type\n'
        '- @w: would be replaced by the wildcard, usually the sdkconfig\n'
        '- @n: would be replaced by the app name\n'
        '- @f: would be replaced by the escaped app path (replaced "/" to "_")\n'
        '- @v: Would be replaced by the ESP-IDF version like `5_3_0`\n'
        '- @i: would be replaced by the build index (only available in `build` command)\n'
        '- @p: would be replaced by the parallel index (only available in `build` command)',
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    actions = parser.add_subparsers(dest='action')

    common_args = argparse.ArgumentParser(add_help=False)
    common_args.add_argument(
        '-c',
        '--config-file',
        help='Path to the default configuration file, toml file',
    )

    common_args.add_argument('-p', '--paths', nargs='+', help='One or more paths to look for apps')
    common_args.add_argument('-t', '--target', help='filter apps by given target')
    common_args.add_argument(
        '--build-system', default='cmake', choices=['cmake', 'make'], help='filter apps by given build system'
    )
    common_args.add_argument(
        '--recursive',
        action='store_true',
        help='Look for apps in the specified paths recursively',
    )
    common_args.add_argument('--exclude', nargs='+', help='Ignore specified path (if --recursive is given)')
    common_args.add_argument(
        '--work-dir',
        help='If set, the app is first copied into the specified directory, and then built. '
        'If not set, the work directory is the directory of the app. Can expand placeholders',
    )
    common_args.add_argument(
        '--build-dir',
        default='build',
        help='If set, specifies the build directory name. Can be either a name relative to the work directory, '
        'or an absolute path. Can expand placeholders',
    )
    common_args.add_argument(
        '--build-log',
        help='Relative to build dir. The build log will be written to this file instead of sys.stdout if specified. '
        'Can expand placeholders',
    )
    common_args.add_argument(
        '--size-file',
        help='Relative to build dir. The size json will be written to this file if specified. Can expand placeholders',
    )
    common_args.add_argument(
        '--config',
        nargs='+',
        help='Adds configurations (sdkconfig file names) to build. '
        'This can either be FILENAME[=NAME] or FILEPATTERN. FILENAME is the name of the sdkconfig file, '
        'relative to the project directory, to be used. Optional NAME can be specified, '
        'which can be used as a name of this configuration. FILEPATTERN is the name of '
        'the sdkconfig file, relative to the project directory, with at most one wildcard. '
        'The part captured by the wildcard is used as the name of the configuration',
    )

    common_args.add_argument(
        '--override-sdkconfig-items',
        nargs='?',
        type=str,
        help='The --override-sdkconfig-items option is a comma-separated list '
        'that permits the overriding of specific configuration items defined '
        "in the SDK's sdkconfig file and Kconfig using a command-line argument. "
        'The sdkconfig items specified here override the same sdkconfig '
        'item defined in the --override-sdkconfig-files, if exists.',
    )
    common_args.add_argument(
        '--override-sdkconfig-files',
        nargs='?',
        type=str,
        help='"The --override-sdkconfig-files option is a comma-separated list, '
        'which provides an alternative (alt: --override-sdkconfig-items) '
        'approach for overriding SDK configuration items. '
        'The filepath may be global or relative to the root.',
    )
    common_args.add_argument(
        '--sdkconfig-defaults',
        help='semicolon-separated string, pass to idf.py -DSDKCONFIG_DEFAULTS if specified, also could be set via '
        'environment variables "SDKCONFIG_DEFAULTS"',
    )
    common_args.add_argument(
        '-v',
        '--verbose',
        default=0,
        action='count',
        help='Increase the logging level of the whole process. Can be specified multiple times. '
        'By default set to WARNING level. '
        'Specify once to set to INFO level. '
        'Specify twice or more to set to DEBUG level',
    )
    common_args.add_argument(
        '--log-file',
        help='Write the log to the specified file, instead of stderr',
    )
    common_args.add_argument(
        '--check-warnings', action='store_true', help='If set, fail the build if warnings are found'
    )

    common_args.add_argument(
        '--manifest-file',
        nargs='+',
        help='Manifest files which specify the build test rules of the apps',
    )
    common_args.add_argument(
        '--manifest-rootpath',
        help='Root directory for calculating the realpath of the relative path defined in the manifest files. '
        'Would use the current directory if not set',
    )
    common_args.add_argument(
        '--check-manifest-rules',
        action='store_true',
        help='Exit with error if any of the manifest rules does not exist on your filesystem',
    )
    common_args.add_argument(
        '--enable-preview-targets',
        action='store_true',
        help='Build the apps with all targets in the current ESP-IDF branch, '
        'including preview targets, when the app supports the target.',
    )
    common_args.add_argument(
        '--default-build-targets',
        nargs='+',
        help='space-separated list of string which specifies the targets for building the apps. '
        'If provided, the apps will be built only with the specified targets '
        'when the app supports the target.',
    )
    common_args.add_argument(
        '--modified-components',
        type=semicolon_separated_str_to_list,
        help='semicolon-separated string which specifies the modified components. '
        'app with `depends_components` set in the corresponding manifest files would only be built '
        'if depends on any of the specified components. '
        'If set to "", the value would be considered as None. '
        'If set to ";", the value would be considered as an empty list',
    )
    common_args.add_argument(
        '--modified-files',
        type=semicolon_separated_str_to_list,
        help='semicolon-separated string which specifies the modified files. '
        'app with `depends_filepatterns` set in the corresponding manifest files would only be built '
        'if any of the specified file pattern matches any of the specified modified files. '
        'If set to "", the value would be considered as None. '
        'If set to ";", the value would be considered as an empty list',
    )
    common_args.add_argument(
        '-ic',
        '--ignore-app-dependencies-components',
        type=semicolon_separated_str_to_list,
        help='semicolon-separated string which specifies the modified components used for '
        'ignoring checking the app dependencies. '
        'The `depends_components` and `depends_filepatterns` set in the manifest files will be ignored when any of the '
        'specified components matches any of the modified components. '
        'Must be used together with --modified-components. '
        'If set to "", the value would be considered as None. '
        'If set to ";", the value would be considered as an empty list',
    )
    common_args.add_argument(
        '-if',
        '--ignore-app-dependencies-filepatterns',
        type=semicolon_separated_str_to_list,
        help='semicolon-separated string which specifies the file patterns used for '
        'ignoring checking the app dependencies. '
        'The `depends_components` and `depends_filepatterns` set in the manifest files will be ignored when any of the '
        'specified file patterns matches any of the modified files. '
        'Must be used together with --modified-files. '
        'If set to "", the value would be considered as None. '
        'If set to ";", the value would be considered as an empty list',
    )

    common_args.add_argument(
        '--no-color',
        action='store_true',
        help='enable colored output by default on UNIX-like systems. enable this flag to make the logs uncolored.',
    )

    find_parser = actions.add_parser(
        'find',
        help='Find the buildable applications. Run `idf-build-apps find --help` for more information on a command.',
        description='Find the buildable applications in the given path or paths for specified chips. '
        '`--path` and `--target` options must be provided. '
        'By default, print the found apps in stdout. '
        'To find apps for all chips use the `--target` option with the `all` argument.',
        parents=[common_args],
        formatter_class=IdfBuildAppsCliFormatter,
    )
    find_parser.add_argument('-o', '--output', help='Print the found apps to the specified file instead of stdout')
    find_parser.add_argument(
        '--output-format',
        choices=['raw', 'json'],
        default='raw',
        help='Output format. In "raw" format, each line is a valid json that represents the app. '
        'In "json" format, the whole file is a JSON file of a list of apps.',
    )
    find_parser.add_argument(
        '--include-all-apps',
        action='store_true',
        help='Include skipped and disabled apps. By default only apps that should be built.',
    )

    build_parser = actions.add_parser(
        'build',
        help='Build the found applications. Run `idf-build-apps build --help` for more information on a command.',
        description='Build the application in the given path or paths for specified chips. '
        '`--path` and `--target` options must be provided.',
        parents=[common_args],
        formatter_class=IdfBuildAppsCliFormatter,
    )
    build_parser.add_argument(
        '--build-verbose',
        action='store_true',
        help='Enable verbose output of the build system',
    )
    build_parser.add_argument(
        '--parallel-count',
        default=1,
        type=int,
        help="Number of parallel build jobs. Note that this script doesn't start all jobs simultaneously. "
        'It needs to be executed multiple times with same value of --parallel-count and '
        'different values of --parallel-index',
    )
    build_parser.add_argument(
        '--parallel-index',
        default=1,
        type=int,
        help='Index (1-based) of the job, out of the number specified by --parallel-count',
    )
    build_parser.add_argument(
        '--dry-run',
        action='store_true',
        help="Don't actually build, only print the build commands",
    )
    build_parser.add_argument(
        '--keep-going',
        action='store_true',
        help="Don't exit immediately when a build fails",
    )
    build_parser.add_argument(
        '--no-preserve',
        action='store_true',
        help="Don't preserve the build directory after a successful build",
    )
    build_parser.add_argument(
        '--collect-size-info',
        help='write size info json file while building into the specified file. each line is a json object. '
        'Can expand placeholder @p',
    )
    build_parser.add_argument(
        '--collect-app-info',
        help='write app info json file while building into the specified file. each line is a json object. '
        'Can expand placeholder @p',
    )
    build_parser.add_argument(
        '--ignore-warning-str',
        nargs='+',
        help='Ignore the warning string that match the specified regex in the build output',
    )
    build_parser.add_argument(
        '--ignore-warning-file',
        type=argparse.FileType('r'),
        help='Ignore the warning strings in the specified file. Each line should be a regex string',
    )
    build_parser.add_argument(
        '--copy-sdkconfig',
        action='store_true',
        help='Copy the sdkconfig file to the build directory',
    )
    build_parser.add_argument(
        '--junitxml',
        help='Path to the junitxml file. If specified, the junitxml file will be generated. Can expand placeholder @p',
    )

    completions_parser = actions.add_parser(
        'completions',
        help='Add the autocompletion activation script to the shell rc file. '
        'Run `idf-build-apps completions --help` for more information on a command.',
        description='Without `--activate` option print instructions for manual activation. '
        'With the `--activate` option, add the autocompletion activation script to the shell rc file '
        'for bash, zsh, or fish. Other shells are not supported.'
        'The `--shell` option is used only with the `--activate` option, '
        'if provided, add the autocompletion activation script to the given shell; '
        'without this argument, will detect shell type automatically. '
        'May need to restart or re-login for the autocompletion to start working',
        formatter_class=IdfBuildAppsCliFormatter,
    )
    completions_parser.add_argument(
        '-a', '--activate', action='store_true', help='Activate autocompletion automatically and permanently. '
    )
    completions_parser.add_argument(
        '-s',
        '--shell',
        choices=['bash', 'zsh', 'fish'],
        help='Specify the shell type for the autocomplite activation script. ',
    )

    return parser


def handle_completions(args: argparse.Namespace) -> None:
    if args.activate:
        if not args.shell:
            args.shell = 'auto'
        activate_completions(args.shell)
    elif not args.activate and args.shell:
        raise AutocompleteActivationError('The --shell option can only be used with the --activate option.')
    else:
        print(completion_instructions)


def validate_args(parser: argparse.ArgumentParser, args: argparse.Namespace) -> None:
    # validate cli subcommands
    if args.action not in ['find', 'build', 'completions']:
        parser.print_help()
        raise InvalidCommand('subcommand is required. {find, build, completions}')

    if not args.paths:
        raise InvalidCommand(
            'Must specify at least one path to search for the apps ' 'with CLI option "-p <path>" or "--path <path>"'
        )

    if not args.target:
        raise InvalidCommand(
            'Must specify current build target with CLI option "-t <target>" or "--target <target>". '
            '(choices: [{}]'.format(','.join([*ALL_TARGETS, 'all']))
        )

    default_build_targets = []
    if args.default_build_targets:
        for target in args.default_build_targets:
            if target not in ALL_TARGETS:
                LOGGER.warning(
                    f'Ignoring... Unrecognizable target {target} specified with "--default-build-targets". '
                    f'Current ESP-IDF available targets: {ALL_TARGETS}'
                )
            elif target not in default_build_targets:
                default_build_targets.append(target)
        args.default_build_targets = default_build_targets
    elif args.enable_preview_targets:
        args.default_build_targets = deepcopy(ALL_TARGETS)

    if args.ignore_app_dependencies_components is not None:
        if args.modified_components is None:
            raise InvalidCommand('Must specify "--ignore-app-dependencies-components" with "--modified-components", ')

    if args.ignore_app_dependencies_filepatterns is not None:
        if args.modified_files is None:
            raise InvalidCommand('Must specify "--ignore-app-dependencies-filepatterns" with "--modified-files", ')


def apply_config_args(args: argparse.Namespace) -> None:
    # support toml config file
    config_dict = get_valid_config(custom_path=args.config_file)
    if config_dict:
        for k, v in config_dict.items():
            setattr(args, k, v)

    setup_logging(args.verbose, args.log_file, not args.no_color)


def main():
    parser = get_parser()
    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    if args.action == 'completions':
        handle_completions(args)
        sys.exit(0)

    apply_config_args(args)
    validate_args(parser, args)

    SESSION_ARGS.set(args)

    if args.action == 'build':
        args.output = None  # build action doesn't support output option

    kwargs = {
        'build_system': args.build_system,
        'recursive': args.recursive,
        'exclude_list': args.exclude or [],
        'work_dir': args.work_dir,
        'build_dir': args.build_dir or 'build',
        'config_rules_str': args.config,
        'build_log_filename': args.build_log,
        'size_json_filename': args.size_file,
        'check_warnings': args.check_warnings,
        'manifest_rootpath': args.manifest_rootpath,
        'manifest_files': args.manifest_file,
        'check_manifest_rules': args.check_manifest_rules,
        'default_build_targets': args.default_build_targets,
        'modified_components': args.modified_components,
        'modified_files': args.modified_files,
        'ignore_app_dependencies_components': args.ignore_app_dependencies_components,
        'ignore_app_dependencies_filepatterns': args.ignore_app_dependencies_filepatterns,
        'sdkconfig_defaults': args.sdkconfig_defaults,
    }
    # only useful in find
    if args.action == 'find' and args.include_all_apps:
        kwargs['include_skipped_apps'] = True
        kwargs['include_disabled_apps'] = True

    # real call starts here
    apps = find_apps(args.paths, args.target, **kwargs)

    if args.action == 'find':
        if args.output:
            os.makedirs(os.path.dirname(os.path.realpath(args.output)), exist_ok=True)
            if args.output.endswith('.json'):
                LOGGER.info('Detecting output file ends with ".json", writing json file.')
                args.output_format = 'json'

            with open(args.output, 'w') as fw:
                if args.output_format == 'raw':
                    for app in apps:
                        fw.write(app.to_json() + '\n')
                elif args.output_format == 'json':
                    fw.write(json.dumps([app.model_dump() for app in apps], indent=2))
                else:
                    raise ValueError(f'Output format {args.output_format} is not supported.')
        else:
            for app in apps:
                print(app)

        sys.exit(0)

    if args.no_preserve:
        for app in apps:
            app.preserve = False

    res = build_apps(
        apps,
        build_verbose=args.build_verbose,
        parallel_count=args.parallel_count,
        parallel_index=args.parallel_index,
        dry_run=args.dry_run,
        keep_going=args.keep_going,
        collect_size_info=args.collect_size_info,
        collect_app_info=args.collect_app_info,
        ignore_warning_strs=args.ignore_warning_str,
        ignore_warning_file=args.ignore_warning_file,
        copy_sdkconfig=args.copy_sdkconfig,
        manifest_rootpath=args.manifest_rootpath,
        modified_components=args.modified_components,
        modified_files=args.modified_files,
        ignore_app_dependencies_components=args.ignore_app_dependencies_components,
        ignore_app_dependencies_filepatterns=args.ignore_app_dependencies_filepatterns,
        junitxml=args.junitxml,
    )

    built_apps = [app for app in apps if app.build_status == BuildStatus.SUCCESS]
    if built_apps:
        print('Successfully built the following apps:')
        for app in built_apps:
            print(f'  {app}')

    skipped_apps = [app for app in apps if app.build_status == BuildStatus.SKIPPED]
    if skipped_apps:
        print('Skipped building the following apps:')
        for app in skipped_apps:
            print(f'  {app}')

    failed_apps = [app for app in apps if app.build_status == BuildStatus.FAILED]
    if failed_apps:
        print('Failed building the following apps:')
        for app in failed_apps:
            print(f'  {app}')

    sys.exit(res)


def json_to_app(json_str: str, extra_classes: t.Optional[t.List[t.Type[App]]] = None) -> App:
    """
    Deserialize json string to App object

    .. note::

        You can pass extra_cls to support custom App class. A custom App class must be a subclass of App, and have a
        different value of `build_system`. For example, a custom CMake app

        >>> class CustomApp(CMakeApp):
        >>>    build_system: Literal['custom_cmake'] = 'custom_cmake'

        Then you can pass the CustomApp class to the `extra_cls` argument

        >>> json_str = CustomApp('.', 'esp32').to_json()
        >>> json_to_app(json_str, extra_classes=[CustomApp])

    :param json_str: json string
    :param extra_classes: extra App class
    :return: App object
    """
    types = [App, CMakeApp, MakeApp]
    if extra_classes:
        types.extend(extra_classes)

    custom_deserializer = create_model(
        '_CustomDeserializer',
        app=(t.Union[tuple(types)], Field(discriminator='build_system')),
        __base__=AppDeserializer,
    )

    return custom_deserializer.from_json(json_str)



# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import operator
import os
from ast import (
    literal_eval,
)
from typing import (
    Any,
)

from packaging.version import (
    Version,
)
from pyparsing import (
    Keyword,
    Literal,
    ParseResults,
    QuotedString,
    Suppress,
    Word,
    alphas,
    delimitedList,
    hexnums,
    infixNotation,
    nums,
    opAssoc,
)

from ..constants import (
    IDF_VERSION,
    IDF_VERSION_MAJOR,
    IDF_VERSION_MINOR,
    IDF_VERSION_PATCH,
)
from ..utils import (
    InvalidInput,
    to_version,
)
from .soc_header import (
    SOC_HEADERS,
)


class Stmt:
    """
    Statement
    """

    def get_value(self, target: str, config_name: str) -> Any:
        """
        Lazy calculated. All subclasses of `Stmt` should implement this function.

        :param target: ESP-IDF target
        :type target: str
        :param config_name: config name
        :type target: str
        :return: the value of the statement
        """
        raise NotImplementedError('Please implement this function in sub classes')


class ChipAttr(Stmt):
    """
    Attributes defined in SOC Header Files and other keywords as followed:

    - IDF_TARGET: target
    - INCLUDE_DEFAULT: take the default build targets into account or not
    - IDF_VERSION_MAJOR: major version of ESP-IDF
    - IDF_VERSION_MINOR: minor version of ESP-IDF
    - IDF_VERSION_PATCH: patch version of ESP-IDF
    - CONFIG_NAME: config name defined in the config rules
    """

    def __init__(self, t: ParseResults):
        self.attr: str = t[0]

    def get_value(self, target: str, config_name: str) -> Any:
        from .manifest import FolderRule  # lazy-load

        if self.attr == 'IDF_TARGET':
            return target

        if self.attr == 'INCLUDE_DEFAULT':
            return 1 if target in FolderRule.DEFAULT_BUILD_TARGETS else 0

        if self.attr == 'IDF_VERSION':
            return IDF_VERSION

        if self.attr == 'IDF_VERSION_MAJOR':
            return IDF_VERSION_MAJOR

        if self.attr == 'IDF_VERSION_MINOR':
            return IDF_VERSION_MINOR

        if self.attr == 'IDF_VERSION_PATCH':
            return IDF_VERSION_PATCH

        if self.attr == 'CONFIG_NAME':
            return config_name

        if self.attr in SOC_HEADERS[target]:
            return SOC_HEADERS[target][self.attr]

        # for non-keyword cap words, check if it is defined in the environment variables
        if self.attr in os.environ:
            return os.environ[self.attr]

        return 0  # default return 0 as false


class Integer(Stmt):
    def __init__(self, t: ParseResults):
        self.expr: str = t[0]

    def get_value(self, target: str, config_name: str) -> Any:  # noqa: ARG002
        return literal_eval(self.expr)


class String(Stmt):
    def __init__(self, t: ParseResults):
        self.expr: str = t[0]

    def get_value(self, target: str, config_name: str) -> Any:  # noqa: ARG002
        return literal_eval(f'"{self.expr}"')  # double quotes is swallowed by QuotedString


class List_(Stmt):
    def __init__(self, t: ParseResults):
        self.expr = t

    def get_value(self, target: str, config_name: str) -> Any:
        return [item.get_value(target, config_name) for item in self.expr]


class BoolStmt(Stmt):
    _OP_DICT = {
        '==': operator.eq,
        '!=': operator.ne,
        '>': operator.gt,
        '>=': operator.ge,
        '<': operator.lt,
        '<=': operator.le,
        'not in': lambda x, y: x not in y,
        'in': lambda x, y: x in y,
    }

    def __init__(self, t: ParseResults):
        self.left: Stmt = t[0]
        self.comparison: str = t[1]
        self.right: Stmt = t[2]

    def get_value(self, target: str, config_name: str) -> Any:
        _l = self.left.get_value(target, config_name)
        _r = self.right.get_value(target, config_name)

        if self.comparison not in ['in', 'not in']:
            # will use version comparison if any of the operands is a Version
            if any(isinstance(x, Version) for x in [_l, _r]):
                _l = to_version(_l)
                _r = to_version(_r)
        else:
            # use str for "in" and "not in" operator
            if isinstance(_l, Version):
                _l = str(_l)
            if isinstance(_r, Version):
                _r = str(_r)

        if self.comparison in self._OP_DICT:
            return self._OP_DICT[self.comparison](_l, _r)

        raise InvalidInput(f'Unsupported comparison operator: "{self.comparison}"')


class BoolExpr(Stmt):
    pass


class BoolAnd(BoolExpr):
    def __init__(self, t: ParseResults):
        self.left: BoolStmt = t[0][0]
        self.right: BoolStmt = t[0][2]

    def get_value(self, target: str, config_name: str) -> Any:
        return self.left.get_value(target, config_name) and self.right.get_value(target, config_name)


class BoolOr(BoolExpr):
    def __init__(self, t: ParseResults):
        self.left: BoolStmt = t[0][0]
        self.right: BoolStmt = t[0][2]

    def get_value(self, target: str, config_name: str) -> Any:
        return self.left.get_value(target, config_name) or self.right.get_value(target, config_name)


CAP_WORD = Word(alphas.upper(), nums + alphas.upper() + '_').setParseAction(ChipAttr)

DECIMAL_NUMBER = Word(nums)
HEX_NUMBER = Literal('0x') + Word(hexnums)
INTEGER = (HEX_NUMBER | DECIMAL_NUMBER).setParseAction(Integer)

STRING = QuotedString('"').setParseAction(String)

LIST = Suppress('[') + delimitedList(INTEGER | STRING).setParseAction(List_) + Suppress(']')

BOOL_OPERAND = CAP_WORD | INTEGER | STRING | LIST

EQ = Keyword('==').setParseAction(lambda t: t[0])
NE = Keyword('!=').setParseAction(lambda t: t[0])
LE = Keyword('<=').setParseAction(lambda t: t[0])
LT = Keyword('<').setParseAction(lambda t: t[0])
GE = Keyword('>=').setParseAction(lambda t: t[0])
GT = Keyword('>').setParseAction(lambda t: t[0])
NOT_IN = Keyword('not in').setParseAction(lambda t: t[0])
IN = Keyword('in').setParseAction(lambda t: t[0])

BOOL_STMT = BOOL_OPERAND + (EQ | NE | LE | LT | GE | GT | NOT_IN | IN) + BOOL_OPERAND
BOOL_STMT.setParseAction(BoolStmt)

AND = Keyword('and')
OR = Keyword('or')

BOOL_EXPR = infixNotation(
    BOOL_STMT,
    [
        (AND, 2, opAssoc.LEFT, BoolAnd),
        (OR, 2, opAssoc.LEFT, BoolOr),
    ],
)



# SPDX-FileCopyrightText: 2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import typing as t

from pydantic import (
    computed_field,
)

from .utils import (
    BaseModel,
)


class BuildAppsArgs(BaseModel):
    PARALLEL_INDEX_PLACEHOLDER: t.ClassVar[str] = '@p'  # replace it with the parallel index

    parallel_index: int = 1
    parallel_count: int = 1

    _junitxml: t.Optional[str] = None
    _collect_app_info: t.Optional[str] = None
    _collect_size_info: t.Optional[str] = None

    def __init__(
        self,
        *,
        collect_app_info: t.Optional[str] = None,
        collect_size_info: t.Optional[str] = None,
        junitxml: t.Optional[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self._junitxml = junitxml
        self._collect_app_info = collect_app_info
        self._collect_size_info = collect_size_info

    @computed_field  # type: ignore
    @property
    def collect_app_info(self) -> t.Optional[str]:
        if self._collect_app_info:
            return self.expand(self._collect_app_info)

        return None

    @computed_field  # type: ignore
    @property
    def collect_size_info(self) -> t.Optional[str]:
        if self._collect_size_info:
            return self.expand(self._collect_size_info)

        return None

    @computed_field  # type: ignore
    @property
    def junitxml(self) -> t.Optional[str]:
        if self._junitxml:
            return self.expand(self._junitxml)

        return None

    def expand(self, path):
        return path.replace(self.PARALLEL_INDEX_PLACEHOLDER, str(self.parallel_index))



# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import tempfile
from pathlib import (
    Path,
)

import pytest
from conftest import (
    create_project,
)

from idf_build_apps.constants import (
    DEFAULT_SDKCONFIG,
    IDF_PATH,
    BuildStatus,
)
from idf_build_apps.main import (
    find_apps,
)


class TestFindWithManifest:
    def test_manifest_rootpath_chdir(self):
        test_dir = Path(IDF_PATH) / 'examples' / 'get-started'

        yaml_file = test_dir / 'test.yml'
        yaml_file.write_text(
            """
examples/get-started:
    enable:
        - if: IDF_TARGET != "esp32"
""",
            encoding='utf8',
        )

        os.chdir(IDF_PATH)
        assert not find_apps(str(test_dir), 'esp32', recursive=True, manifest_files=str(yaml_file))

        # manifest folder invalid
        os.chdir(test_dir)
        with pytest.warns(
            UserWarning, match=f'Folder "{IDF_PATH}/examples/get-started/examples/get-started" does not exist'
        ):
            assert find_apps(str(test_dir), 'esp32', recursive=True, manifest_files=str(yaml_file))

    def test_manifest_rootpath_specified(self):
        test_dir = Path(IDF_PATH) / 'examples' / 'get-started'

        yaml_file = test_dir / 'test.yml'
        yaml_file.write_text(
            """
get-started:
    enable:
        - if: IDF_TARGET != "esp32"
""",
            encoding='utf8',
        )
        with pytest.warns(UserWarning, match=f'Folder "{IDF_PATH}/get-started" does not exist.'):
            assert find_apps(
                str(test_dir), 'esp32', recursive=True, manifest_files=str(yaml_file), manifest_rootpath=str(IDF_PATH)
            )

        assert not find_apps(
            str(test_dir),
            'esp32',
            recursive=True,
            manifest_files=str(yaml_file),
            manifest_rootpath=os.path.join(IDF_PATH, 'examples'),
        )

    def test_keyword_idf_target(self, tmpdir):
        test_dir = os.path.join(IDF_PATH, 'examples')
        apps = find_apps(test_dir, 'esp32', recursive=True)
        assert apps

        yaml_file = tmpdir / 'test.yml'
        yaml_file.write_text(
            f"""
{test_dir}:
    enable:
        - if: IDF_TARGET == "esp32s2"
""",
            encoding='utf8',
        )
        filtered_apps = find_apps(test_dir, 'esp32', recursive=True, manifest_files=yaml_file)
        assert not filtered_apps
        assert filtered_apps != apps

    def test_keyword_idf_version(self):
        test_dir = Path(IDF_PATH) / 'examples' / 'get-started'
        apps = find_apps(str(test_dir), 'esp32', recursive=True)
        assert apps

        yaml_file = test_dir / 'test.yml'
        yaml_file.write_text(
            """
examples/get-started:
    enable:
        - if: IDF_VERSION_MAJOR > 0 and IDF_VERSION_MINOR < 999 and IDF_VERSION_PATCH in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
""",
            encoding='utf8',
        )
        assert (
            find_apps(str(test_dir), 'esp32', recursive=True, manifest_rootpath=IDF_PATH, manifest_files=str(yaml_file))
            == apps
        )

    def test_include_disabled_apps(self, tmpdir):
        test_dir = Path(IDF_PATH) / 'examples' / 'get-started'
        yaml_file = tmpdir / 'test.yml'
        yaml_file.write_text(
            f"""
        {test_dir}:
            enable:
                - if: IDF_TARGET == "esp32s2"
        """,
            encoding='utf8',
        )

        test_dir = Path(IDF_PATH) / 'examples' / 'get-started'
        apps = find_apps(
            str(test_dir), 'esp32', recursive=True, manifest_files=str(yaml_file), include_disabled_apps=False
        )
        assert not apps

        apps = find_apps(
            str(test_dir), 'esp32', recursive=True, manifest_files=str(yaml_file), include_disabled_apps=True
        )
        assert apps
        for app in apps:
            assert app.build_status == BuildStatus.DISABLED
            app.build()
            assert app.build_comment == 'Build disabled. Skipping...'


class TestFindWithModifiedFilesComponents:
    @pytest.mark.parametrize(
        'modified_components, could_find_apps',
        [
            (None, True),
            ([], False),
            ('fake', False),
            ('soc', True),
            (['soc', 'fake'], True),
        ],
    )
    def test_with_depends_and_modified_components(self, tmpdir, modified_components, could_find_apps):
        test_dir = str(Path(IDF_PATH) / 'examples')
        apps = find_apps(test_dir, 'esp32', recursive=True)
        assert apps

        yaml_file = tmpdir / 'test.yml'
        yaml_file.write_text(
            f"""
{test_dir}:
    depends_components:
        - freertos
        - soc
""",
            encoding='utf8',
        )

        filtered_apps = find_apps(
            test_dir,
            'esp32',
            recursive=True,
            manifest_files=yaml_file,
            modified_components=modified_components,
        )
        if could_find_apps:
            assert filtered_apps == apps
        else:
            assert not filtered_apps

    @pytest.mark.parametrize(
        'modified_files, could_find_apps',
        [
            ('/foo', False),
            (os.path.join(IDF_PATH, 'examples', 'README.md'), False),
            ([os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'README.md')], False),
            (
                [
                    os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'README.md'),
                    os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'main', 'hello_world_main.c'),
                ],
                True,
            ),
        ],
    )
    def test_with_depends_components_but_modified(self, tmp_path, modified_files, could_find_apps):
        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')
        apps = find_apps(test_dir, 'esp32', recursive=True)
        assert apps

        yaml_file = tmp_path / 'test.yml'
        yaml_file.write_text(
            f"""
{test_dir}:
    depends_components:
        - soc
""",
            encoding='utf8',
        )

        filtered_apps = find_apps(
            test_dir,
            'esp32',
            recursive=True,
            manifest_files=str(yaml_file),
            modified_components=[],
            modified_files=modified_files,
        )
        if could_find_apps:
            assert filtered_apps == apps
        else:
            assert not filtered_apps

    @pytest.mark.parametrize('suffix', ['/', '/   '])
    def test_app_name_is_not_null(self, suffix):
        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world') + suffix
        apps = find_apps(test_dir, 'esp32', recursive=True)
        for app in apps:
            assert app.name == 'hello_world'

        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world') + suffix
        apps = find_apps(test_dir, 'esp32', recursive=False)
        for app in apps:
            assert app.name == 'hello_world'

        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started') + suffix

        apps = find_apps(test_dir, 'esp32', recursive=True)
        for app in apps:
            assert len(app.name.strip()) != 0

    @pytest.mark.parametrize(
        'modified_components, modified_files, could_find_apps',
        [
            ([], os.path.join(IDF_PATH, 'examples', 'README.md'), (True, False)),
            (None, [os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'README.md')], (True, True)),
            (
                [],
                [
                    os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'README.md'),
                    os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'main', 'hello_world_main.c'),
                ],
                (True, True),
            ),
        ],
    )
    def test_with_depends_components_and_filepatterns(
        self, tmp_path, modified_components, modified_files, could_find_apps
    ):
        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')
        apps = find_apps(test_dir, 'esp32', recursive=True)
        assert apps

        yaml_file = tmp_path / 'test.yml'
        yaml_file.write_text(
            f"""
{test_dir}:
    depends_filepatterns:
        - examples/get-started/hello_world/**
        - examples/foo/**
""",
            encoding='utf8',
        )

        filtered_apps = find_apps(
            test_dir,
            'esp32',
            recursive=True,
            manifest_files=str(yaml_file),
            modified_components=modified_components,
        )
        if could_find_apps[0]:
            assert {app.app_dir for app in filtered_apps} == {app.app_dir for app in filtered_apps}
        else:
            assert not filtered_apps

        yaml_file.write_text(
            f"""
{test_dir}:
    depends_components:
        - foo
    depends_filepatterns:
        - examples/get-started/hello_world/**
        - examples/foo/**
""",
            encoding='utf8',
        )

        filtered_apps = find_apps(
            test_dir,
            'esp32',
            recursive=True,
            manifest_rootpath=IDF_PATH,
            manifest_files=str(yaml_file),
            modified_components=modified_components,
            modified_files=modified_files,
        )
        if could_find_apps[1]:
            assert {app.app_dir for app in filtered_apps}
        else:
            assert not filtered_apps

    @pytest.mark.parametrize(
        'modified_files',
        [
            'examples/wifi/getting_started/softAP/main/softap_example_main.c',
            'examples/wifi/getting_started/station/main/station_example_main.c',
        ],
    )
    def test_depends_filepatterns_only_on_parent_dir(self, tmp_path, modified_files):
        test_dir = os.path.join(IDF_PATH, 'examples', 'wifi', 'getting_started')

        apps = find_apps(test_dir, 'esp32', recursive=True)
        assert apps

        yaml_file = tmp_path / 'test.yml'
        with open(yaml_file, 'w') as fw:
            fw.write(
                """
examples/wifi/getting_started:
    depends_filepatterns:
        - examples/wifi/getting_started/**/*
""",
            )

        filtered_apps = find_apps(
            test_dir,
            'esp32',
            recursive=True,
            modified_components=[],
            modified_files=[modified_files],
            manifest_rootpath=IDF_PATH,
            manifest_files=[yaml_file],
        )
        assert len(filtered_apps) == len(apps)

    @pytest.mark.parametrize(
        'modified_files, could_find_apps',
        [
            (None, True),
            (os.path.join(IDF_PATH, 'examples', 'README.md'), True),
            ([os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'README.md')], True),
            (
                [
                    os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'README.md'),
                    os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'main', 'hello_world_main.c'),
                ],
                True,
            ),
            ([os.path.join(IDF_PATH, 'examples', 'a.c')], True),
        ],
    )
    def test_with_filepattern_but_calculate_component_later(self, modified_files, could_find_apps):
        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')
        apps = find_apps(test_dir, 'esp32', recursive=True)
        assert apps

        filtered_apps = find_apps(
            test_dir,
            'esp32',
            recursive=True,
            modified_files=modified_files,
        )
        if could_find_apps:
            assert {app.app_dir for app in filtered_apps} == {app.app_dir for app in apps}
        else:
            assert not filtered_apps

    def test_with_skipped_but_included(self, tmp_path):
        create_project('foo', tmp_path)

        apps = find_apps(
            str(tmp_path / 'foo'), 'esp32', modified_components=[], modified_files=[], include_skipped_apps=True
        )
        assert len(apps) == 1
        assert apps[0].build_status == BuildStatus.SKIPPED

        apps[0].build()
        assert apps[0].build_status == BuildStatus.SKIPPED


class TestFindWithSdkconfigFiles:
    def test_with_sdkconfig_defaults_idf_target(self):
        test_dir = os.path.join(IDF_PATH, 'examples')
        apps = find_apps(test_dir, 'esp32', recursive=True)
        assert apps

        # write the first app without sdkconfig.defaults with CONFIG_IDF_TARGET="linux" in sdkconfig.defaults
        _app = None
        _default_sdkconfig_path = None
        for app in apps:
            default_sdkconfig_path = os.path.join(app.app_dir, DEFAULT_SDKCONFIG)
            if not os.path.isfile(default_sdkconfig_path):
                with open(default_sdkconfig_path, 'w') as fw:
                    fw.write('CONFIG_IDF_TARGET="linux"')
                    logging.info('Created temp %s %s', DEFAULT_SDKCONFIG, default_sdkconfig_path)
                _app = app
                _default_sdkconfig_path = default_sdkconfig_path
                break
        else:
            raise ValueError(f'no app without {DEFAULT_SDKCONFIG}')

        try:
            filtered_apps = find_apps(test_dir, 'esp32', recursive=True)
            assert set(apps) - set(filtered_apps) == {_app}
        finally:
            try:
                os.remove(_default_sdkconfig_path)
                logging.info('Removed temp %s %s', DEFAULT_SDKCONFIG, _default_sdkconfig_path)
            except:  # noqa
                pass

    def test_with_sdkconfig_defaults_idf_target_but_disabled(self, tmp_path):
        manifest_file = tmp_path / 'manifest.yml'
        manifest_file.write_text(
            f"""
{IDF_PATH}/examples:
    disable:
       - if: IDF_TARGET == "esp32s2"
"""
        )

        sdkconfig_defaults = tmp_path / 'sdkconfig.defaults'
        sdkconfig_defaults.write_text('CONFIG_IDF_TARGET="esp32s2"')

        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')
        assert not find_apps(
            test_dir,
            'esp32s2',
            recursive=True,
            sdkconfig_defaults=str(sdkconfig_defaults),
            manifest_files=[str(manifest_file)],
        )

    def test_with_config_rules(self, tmp_path, monkeypatch):
        create_project('test1', tmp_path)

        (tmp_path / 'test1' / 'sdkconfig.defaults').touch()
        (tmp_path / 'test1' / 'sdkconfig.defaults_new').touch()
        (tmp_path / 'test1' / 'sdkconfig.ci.foo').touch()

        apps = find_apps(str(tmp_path / 'test1'), 'esp32', recursive=True, config_rules_str='sdkconfig.ci.*=')
        assert len(apps) == 1
        assert apps[0].sdkconfig_files == [
            str(tmp_path / 'test1' / 'sdkconfig.defaults'),
            str(tmp_path / 'test1' / 'sdkconfig.ci.foo'),
        ]

        monkeypatch.setenv('SDKCONFIG_DEFAULTS', 'sdkconfig.defaults_new')
        apps = find_apps(str(tmp_path / 'test1'), 'esp32', recursive=True, config_rules_str='sdkconfig.ci.*=')
        assert len(apps) == 1
        assert apps[0].sdkconfig_files == [
            str(tmp_path / 'test1' / 'sdkconfig.defaults_new'),
            str(tmp_path / 'test1' / 'sdkconfig.ci.foo'),
        ]

        apps = find_apps(
            str(tmp_path),
            'esp32',
            recursive=True,
            config_rules_str='sdkconfig.ci.*=',  # wrong one
            sdkconfig_defaults='notexists;sdkconfig.defaults_new;sdkconfig.defaults',
        )
        assert len(apps) == 1
        assert apps[0].sdkconfig_files == [
            str(tmp_path / 'test1' / 'sdkconfig.defaults_new'),
            str(tmp_path / 'test1' / 'sdkconfig.defaults'),
            str(tmp_path / 'test1' / 'sdkconfig.ci.foo'),
        ]

    def test_with_sdkconfig_defaults_env_var_expansion(self, tmp_path, monkeypatch):
        create_project('test1', tmp_path)
        (tmp_path / 'test1' / 'sdkconfig.ci.foo').write_text('CONFIG_IDF_TARGET=${TEST_TARGET}', encoding='utf8')
        (tmp_path / 'test1' / 'sdkconfig.ci.bar').write_text('CONFIG_IDF_TARGET=esp32s2', encoding='utf8')
        (tmp_path / 'test1' / 'sdkconfig.ci.baz').write_text('CONFIG_IDF_TARGET=esp32s3', encoding='utf8')

        monkeypatch.setenv('TEST_TARGET', 'esp32')
        apps = find_apps(str(tmp_path), 'esp32', recursive=True, config_rules_str='sdkconfig.ci.*=')
        assert len(apps) == 1
        assert Path(apps[0].sdkconfig_files[0]).parts[-3:] == ('expanded_sdkconfig_files', 'build', 'sdkconfig.ci.foo')

        # test relative paths
        os.chdir(str(tmp_path))
        apps = find_apps('test1', 'esp32', recursive=True, config_rules_str='sdkconfig.ci.*=', build_dir='build_@t_@w')
        assert len(apps) == 1
        assert Path(apps[0].sdkconfig_files[0]).parts[-3:] == (
            'expanded_sdkconfig_files',
            'build_esp32_foo',
            'sdkconfig.ci.foo',
        )

        monkeypatch.setenv('TEST_TARGET', 'esp32s2')
        apps = find_apps('test1', 'esp32', recursive=True, config_rules_str='sdkconfig.ci.*=')
        assert len(apps) == 0

    def test_with_sdkconfig_override(self, tmp_path):
        import idf_build_apps

        create_project('test1', tmp_path)
        (tmp_path / 'test1' / 'sdkconfig.defaults').write_text(
            """
CONFIG_IDF_TARGET=esp32s2
CONFIG_FREERTOS_IDLE_TASK_STACKSIZE=1516
""",
            encoding='utf8',
        )
        (tmp_path / 'sdkconfig.override1').write_text(
            'CONFIG_IDF_TARGET=esp32\nCONFIG_A=5\nCONFIG_B=33', encoding='utf8'
        )
        from collections import (
            namedtuple,
        )

        Args = namedtuple('Args', ['override_sdkconfig_items', 'override_sdkconfig_files'])
        args = Args(
            'CONFIG_IDF_TARGET=esp32c3,CONFIG_FREERTOS_IDLE_TASK_STACKSIZE=1522,CONFIG_A=10,CONFIG_C="test"',
            'sdkconfig.override1',
        )
        idf_build_apps.SESSION_ARGS.set(args, workdir=tmp_path)
        session_args = idf_build_apps.SESSION_ARGS

        assert str(tmp_path) in str(session_args.override_sdkconfig_file_path)
        assert len(session_args.override_sdkconfig_items) == 5
        assert len(set(session_args.override_sdkconfig_items)) == 5

        apps = find_apps(str(tmp_path / 'test1'), 'esp32s2')
        assert len(apps) == 0

        apps = find_apps(str(tmp_path / 'test1'), 'esp32')
        assert len(apps) == 0

        apps = find_apps(str(tmp_path / 'test1'), 'esp32c3')
        assert len(apps) == 1

        with open(apps[0].sdkconfig_files[-1]) as fr:
            lines = fr.readlines()
            assert len(lines) == 5
            for _l in lines:
                assert _l in [
                    'CONFIG_IDF_TARGET=esp32c3\n',
                    'CONFIG_A=10\n',
                    'CONFIG_FREERTOS_IDLE_TASK_STACKSIZE=1522\n',
                    'CONFIG_B=33\n',
                    'CONFIG_C="test"\n',
                ]

    def test_config_name_in_manifest(self, tmp_path):
        create_project('test1', tmp_path)

        (tmp_path / 'test1' / 'sdkconfig.defaults').touch()
        (tmp_path / 'test1' / 'sdkconfig.ci').touch()
        (tmp_path / 'test1' / 'sdkconfig.ci.foo').touch()
        (tmp_path / 'test1' / 'sdkconfig.ci.bar').touch()

        yaml_file = tmp_path / 'test.yml'
        yaml_file.write_text(
            f"""
{tmp_path}:
    enable:
    - if: CONFIG_NAME == "foo" and IDF_TARGET == "esp32"
    - if: CONFIG_NAME == "bar" and IDF_TARGET == "esp32s2"
    - if: CONFIG_NAME == "default" and IDF_TARGET == "esp32s3"
""",
            encoding='utf8',
        )

        apps = find_apps(
            str(tmp_path / 'test1'),
            'esp32',
            recursive=True,
            config_rules_str=['sdkconfig.ci.*=', 'sdkconfig.ci=default'],
            manifest_files=str(yaml_file),
        )
        assert len(apps) == 1
        assert apps[0].sdkconfig_files == [
            str(tmp_path / 'test1' / 'sdkconfig.defaults'),
            str(tmp_path / 'test1' / 'sdkconfig.ci.foo'),
        ]
        apps = find_apps(
            str(tmp_path / 'test1'),
            'esp32s2',
            recursive=True,
            config_rules_str=['sdkconfig.ci.*=', 'sdkconfig.ci=default'],
            manifest_files=str(yaml_file),
        )
        assert len(apps) == 1
        assert apps[0].sdkconfig_files == [
            str(tmp_path / 'test1' / 'sdkconfig.defaults'),
            str(tmp_path / 'test1' / 'sdkconfig.ci.bar'),
        ]
        apps = find_apps(
            str(tmp_path / 'test1'),
            'esp32s3',
            recursive=True,
            config_rules_str=['sdkconfig.ci.*=', 'sdkconfig.ci=default'],
            manifest_files=str(yaml_file),
        )
        assert len(apps) == 1
        assert apps[0].sdkconfig_files == [
            str(tmp_path / 'test1' / 'sdkconfig.defaults'),
            str(tmp_path / 'test1' / 'sdkconfig.ci'),
        ]
        apps = find_apps(
            str(tmp_path / 'test1'),
            'esp32s3',
            recursive=True,
            config_rules_str=['=default'],
            manifest_files=str(yaml_file),
        )
        assert len(apps) == 1
        assert apps[0].sdkconfig_files == [
            str(tmp_path / 'test1' / 'sdkconfig.defaults'),
        ]

    def test_env_var(self, tmp_path, monkeypatch):
        create_project('test1', tmp_path)

        (tmp_path / 'test1' / 'sdkconfig.ci.foo').touch()
        (tmp_path / 'test1' / 'sdkconfig.ci.bar').touch()
        (tmp_path / 'test1' / 'sdkconfig.ci.baz').touch()

        yaml_file = tmp_path / 'test.yml'
        yaml_file.write_text(
            f"""
{tmp_path}:
  enable:
    - if: CONFIG_NAME == "foo" and IDF_TARGET == "esp32"
    - if: CONFIG_NAME == "bar" and IDF_TARGET == "esp32s2"
    - if: TEST_ENV_VAR == "1"
    - if: CONFIG_NAME == "baz" and TEST_ENV_VAR == 0
""",
            encoding='utf8',
        )

        # in case you set it...
        monkeypatch.delenv('CONFIG_NAME', raising=False)
        monkeypatch.delenv('TEST_ENV_VAR', raising=False)

        # CONFIG_NAME should NOT be overridden by env var
        monkeypatch.setenv('CONFIG_NAME', 'bar')
        apps = find_apps(
            str(tmp_path / 'test1'),
            'esp32',
            config_rules_str=['sdkconfig.ci=default', 'sdkconfig.ci.*='],
            manifest_files=str(yaml_file),
        )
        assert len(apps) == 2
        assert apps[0].sdkconfig_files == [
            str(tmp_path / 'test1' / 'sdkconfig.ci.baz'),
        ]
        assert apps[1].sdkconfig_files == [
            str(tmp_path / 'test1' / 'sdkconfig.ci.foo'),
        ]
        monkeypatch.delenv('CONFIG_NAME')

        # env var should be expanded
        monkeypatch.setenv('TEST_ENV_VAR', '1')
        apps = find_apps(
            str(tmp_path / 'test1'),
            'esp32',
            config_rules_str=['sdkconfig.ci=default', 'sdkconfig.ci.*='],
            manifest_files=str(yaml_file),
        )
        assert len(apps) == 3
        monkeypatch.delenv('TEST_ENV_VAR')


@pytest.mark.parametrize(
    'exclude_list, apps_count',
    [
        (['test1'], 3),  # not excluded
        (['folder1/test2'], 2),
        (['folder1'], 1),
        (['folder2/test2'], 2),
    ],
)
def test_find_apps_with_exclude(tmp_path, exclude_list, apps_count):
    (tmp_path / 'folder1').mkdir()
    (tmp_path / 'folder2').mkdir()

    create_project('test1', tmp_path / 'folder1')
    create_project('test2', tmp_path / 'folder1')
    create_project('test2', tmp_path / 'folder2')

    os.chdir(tmp_path)
    apps = find_apps(str(tmp_path), 'esp32', recursive=True, exclude_list=exclude_list)
    assert len(apps) == apps_count

    # or with absolute_path
    exclude_list = [os.path.abspath(x) for x in exclude_list]
    tmp_path = os.path.abspath(tmp_path)
    os.chdir(tempfile.tempdir)
    apps = find_apps(str(tmp_path), 'esp32', recursive=True, exclude_list=exclude_list)
    assert len(apps) == apps_count



# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import shutil
from copy import (
    deepcopy,
)
from xml.etree import (
    ElementTree,
)

import pytest
from conftest import (
    create_project,
)

from idf_build_apps import (
    build_apps,
    find_apps,
)
from idf_build_apps.app import (
    CMakeApp,
)
from idf_build_apps.constants import (
    IDF_PATH,
    BuildStatus,
)


@pytest.mark.skipif(not shutil.which('idf.py'), reason='idf.py not found')
class TestBuild:
    def test_build_hello_world(self, tmpdir, capsys):
        path = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        app = CMakeApp(path, 'esp32', work_dir=str(tmpdir / 'test'))
        app.build()

        captured = capsys.readouterr()
        assert 'Configuring done' in captured.out
        assert 'Project build complete.' in captured.out
        assert app.build_status == BuildStatus.SUCCESS

    @pytest.mark.parametrize(
        'modified_components, check_app_dependencies, build_status',
        [
            (None, True, BuildStatus.SUCCESS),
            ([], True, BuildStatus.SKIPPED),
            ([], False, BuildStatus.SUCCESS),
            ('fake', True, BuildStatus.SKIPPED),
            ('fake', False, BuildStatus.SUCCESS),
            ('soc', True, BuildStatus.SUCCESS),
            ('soc', False, BuildStatus.SUCCESS),
            (['soc', 'fake'], True, BuildStatus.SUCCESS),
        ],
    )
    def test_build_with_modified_components(self, tmpdir, modified_components, check_app_dependencies, build_status):
        path = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        app = CMakeApp(path, 'esp32', work_dir=str(tmpdir / 'test'))
        app.build(
            modified_components=modified_components,
            check_app_dependencies=check_app_dependencies,
        )
        assert app.build_status == build_status

    @pytest.mark.parametrize(
        'modified_files, build_status',
        [
            ('/foo', BuildStatus.SKIPPED),
            (os.path.join(IDF_PATH, 'examples', 'README.md'), BuildStatus.SKIPPED),
            ([os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'README.md')], BuildStatus.SKIPPED),
            (
                [
                    os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'README.md'),
                    os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'main', 'hello_world_main.c'),
                ],
                BuildStatus.SUCCESS,
            ),
        ],
    )
    def test_build_with_modified_files(self, modified_files, build_status):
        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        app = CMakeApp(test_dir, 'esp32')
        app.build(
            modified_components=[],
            modified_files=modified_files,
            check_app_dependencies=True,
        )

        assert app.build_status == build_status

    def test_build_without_modified_components_but_ignored_app_dependency_check(self):
        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        apps = find_apps(
            test_dir,
            'esp32',
            modified_components=[],
            modified_files=['foo.c'],
            ignore_app_dependencies_filepatterns=['foo.c'],
        )

        for app in apps:
            app.build()
            assert app.build_status == BuildStatus.SUCCESS

    def test_build_with_junit_output(self, tmpdir):
        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        apps = [
            CMakeApp(test_dir, 'esp32', build_dir='build_1'),
            CMakeApp(test_dir, 'esp32', build_dir='build_2'),
            CMakeApp(test_dir, 'esp32', build_dir='build_3'),
            CMakeApp(test_dir, 'esp32', build_dir='build_4'),
        ]
        apps[2].build_status = BuildStatus.DISABLED
        apps[3].build_status = BuildStatus.SKIPPED

        build_apps(deepcopy(apps), dry_run=True, junitxml=str(tmpdir / 'test.xml'))

        with open(tmpdir / 'test.xml') as f:
            xml = ElementTree.fromstring(f.read())

        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['tests'] == '0'
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '4'

        for i, testcase in enumerate(test_suite.findall('testcase')):
            assert testcase.attrib['name'] == apps[i].build_path
            assert float(testcase.attrib['time']) > 0
            assert testcase.find('skipped') is not None
            if i in (0, 1):
                assert testcase.find('skipped').attrib['message'] == 'dry run'
            elif i == 2:
                assert testcase.find('skipped').attrib['message'] == 'Build disabled. Skipping...'
            elif i == 3:
                assert testcase.find('skipped').attrib['message'] == 'Build skipped. Skipping...'
            else:
                assert False  # not expected

        build_apps(deepcopy(apps), junitxml=str(tmpdir / 'test.xml'))

        with open(tmpdir / 'test.xml') as f:
            xml = ElementTree.fromstring(f.read())

        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['tests'] == '2'
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '2'

        for i, testcase in enumerate(test_suite.findall('testcase')):
            assert float(testcase.attrib['time']) > 0
            assert testcase.attrib['name'] == apps[i].build_path
            assert testcase.find('error') is None
            assert testcase.find('failure') is None
            if i in (0, 1):
                assert testcase.find('skipped') is None
            elif i == 2:
                assert testcase.find('skipped').attrib['message'] == 'Build disabled. Skipping...'
            elif i == 3:
                assert testcase.find('skipped').attrib['message'] == 'Build skipped. Skipping...'
            else:
                assert False  # not expected

    def test_work_dir_inside_relative_app_dir(self, tmp_path):
        create_project('foo', tmp_path)

        os.chdir(tmp_path / 'foo')
        apps = find_apps(
            '.',
            'esp32',
            work_dir=os.path.join('foo', 'bar'),
        )
        build_apps(apps)

        assert len(apps) == 1
        assert apps[0].build_status == BuildStatus.SUCCESS



# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import typing as t
import warnings

from pyparsing import (
    ParseException,
)

from ..constants import (
    ALL_TARGETS,
    SUPPORTED_TARGETS,
)
from ..utils import (
    InvalidIfClause,
    InvalidManifest,
)
from ..yaml import (
    parse,
)
from .if_parser import (
    BOOL_EXPR,
    BoolStmt,
)

LOGGER = logging.getLogger(__name__)


class IfClause:
    def __init__(self, stmt: str, temporary: bool = False, reason: t.Optional[str] = None) -> None:
        try:
            self.stmt: BoolStmt = BOOL_EXPR.parseString(stmt)[0]
        except ParseException:
            raise InvalidIfClause(f'Invalid if statement: {stmt}')

        self.temporary = temporary
        self.reason = reason

        if self.temporary is True and not self.reason:
            raise InvalidIfClause('"reason" must be set when "temporary: true"')

    def get_value(self, target: str, config_name: str) -> t.Any:
        return self.stmt.get_value(target, config_name)


class SwitchClause:
    def __init__(
        self, if_clauses: t.List[IfClause], contents: t.List[t.List[str]], default_clause: t.List[str]
    ) -> None:
        self.if_clauses = if_clauses
        self.contents = contents
        self.default_clause = default_clause

    def get_value(self, target: str, config_name: str) -> t.Any:
        for if_clause, content in zip(self.if_clauses, self.contents):
            if if_clause.get_value(target, config_name):
                return content
        return self.default_clause


class FolderRule:
    DEFAULT_BUILD_TARGETS = SUPPORTED_TARGETS

    def __init__(
        self,
        folder: str,
        enable: t.Optional[t.List[t.Dict[str, t.Any]]] = None,
        disable: t.Optional[t.List[t.Dict[str, t.Any]]] = None,
        disable_test: t.Optional[t.List[t.Dict[str, t.Any]]] = None,
        depends_components: t.Optional[t.List[t.Union[str, t.Dict[str, t.Any]]]] = None,
        depends_filepatterns: t.Optional[t.List[t.Union[str, t.Dict[str, t.Any]]]] = None,
    ) -> None:
        self.folder = os.path.abspath(folder)

        def _clause_to_if_clause(clause: t.Dict[str, t.Any]) -> IfClause:
            _kwargs = {'stmt': clause['if']}
            if 'temporary' in clause:
                _kwargs['temporary'] = clause['temporary']
            if 'reason' in clause:
                _kwargs['reason'] = clause['reason']
            return IfClause(**_kwargs)

        def _clause_to_switch_or_list(
            statements: t.Optional[t.List[t.Union[str, t.Dict[str, t.Any]]]],
        ) -> t.Union[SwitchClause, t.List[str]]:
            if not statements:
                return []

            switch_statements = []
            str_statements = []
            for statement in statements:
                if isinstance(statement, t.Dict):
                    switch_statements.append(statement)
                else:
                    str_statements.append(statement)

            if switch_statements and str_statements:
                raise InvalidManifest('Current manifest format has to fit either the switch format or the list format.')

            if str_statements:
                return str_statements

            return _clause_to_switch_clause(switch_statements)

        def _clause_to_switch_clause(switch_statements: t.List[t.Dict[str, t.Any]]) -> SwitchClause:
            if_clauses = []
            contents = []
            default_clauses = []
            for statement in switch_statements:
                if 'if' in statement:
                    if_clauses.append(IfClause(stmt=statement['if']))
                    contents.append(statement['content'])
                elif 'default' in statement:
                    default_clauses.extend(statement['default'])
                else:
                    raise InvalidManifest("Only the 'if' and 'default' keywords are supported in switch clause.")

            return SwitchClause(if_clauses, contents, default_clauses)

        self.enable = [_clause_to_if_clause(clause) for clause in enable] if enable else []
        self.disable = [_clause_to_if_clause(clause) for clause in disable] if disable else []
        self.disable_test = [_clause_to_if_clause(clause) for clause in disable_test] if disable_test else []
        self.depends_components = _clause_to_switch_or_list(depends_components)
        self.depends_filepatterns = _clause_to_switch_or_list(depends_filepatterns)

    def __hash__(self) -> int:
        return hash(self.folder)

    def __repr__(self) -> str:
        return f'FolderRule({self.folder})'

    def _enable_build(self, target: str, config_name: str) -> bool:
        if self.enable:
            res = False
            for clause in self.enable:
                if clause.get_value(target, config_name):
                    res = True
                    break
        else:
            res = target in self.DEFAULT_BUILD_TARGETS

        if self.disable:
            for clause in self.disable:
                if clause.get_value(target, config_name):
                    res = False
                    break

        return res

    def _enable_test(
        self, target: str, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> bool:
        res = target in self.enable_build_targets(default_sdkconfig_target, config_name)

        if self.disable or self.disable_test:
            for clause in self.disable + self.disable_test:
                if clause.get_value(target, config_name or ''):
                    res = False
                    break

        return res

    def enable_build_targets(
        self, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        res = []
        for target in ALL_TARGETS:
            if self._enable_build(target, config_name or ''):
                res.append(target)

        if default_sdkconfig_target:
            if default_sdkconfig_target not in res:
                LOGGER.debug(
                    'sdkconfig defined `CONFIG_IDF_TARGET=%s` is not enabled for folder %s. Skip building this App...',
                    default_sdkconfig_target,
                    self.folder,
                )
                return []
            else:
                LOGGER.debug(
                    'sdkconfig defined `CONFIG_IDF_TARGET=%s` overrides the supported targets for folder %s',
                    default_sdkconfig_target,
                    self.folder,
                )
                res = [default_sdkconfig_target]

        return sorted(res)

    def enable_test_targets(
        self, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        res = []
        for target in ALL_TARGETS:
            if self._enable_test(target, default_sdkconfig_target, config_name):
                res.append(target)

        return sorted(res)


class DefaultRule(FolderRule):
    def __init__(self, folder: str) -> None:
        super().__init__(folder)


class Manifest:
    # could be reassigned later
    ROOTPATH = os.curdir
    CHECK_MANIFEST_RULES = False

    def __init__(
        self,
        rules: t.Iterable[FolderRule],
    ) -> None:
        self.rules = sorted(rules, key=lambda x: x.folder)

    @classmethod
    def from_files(cls, paths: t.List[str]) -> 'Manifest':
        # folder, defined_at dict
        _known_folders: t.Dict[str, str] = dict()

        rules: t.List[FolderRule] = []
        for path in paths:
            _manifest = cls.from_file(path)

            for rule in _manifest.rules:
                if rule.folder in _known_folders:
                    msg = f'Folder "{rule.folder}" is already defined in {_known_folders[rule.folder]}'
                    if cls.CHECK_MANIFEST_RULES:
                        raise InvalidManifest(msg)
                    else:
                        warnings.warn(msg)

                _known_folders[rule.folder] = path

            rules.extend(_manifest.rules)

        return Manifest(rules)

    @classmethod
    def from_file(cls, path: str) -> 'Manifest':
        manifest_dict = parse(path)

        rules: t.List[FolderRule] = []
        for folder, folder_rule in manifest_dict.items():
            # not a folder, but a anchor
            if folder.startswith('.'):
                continue

            if not os.path.isabs(folder):
                folder = os.path.join(cls.ROOTPATH, folder)

            if not os.path.exists(folder):
                msg = f'Folder "{folder}" does not exist. Please check your manifest file {path}'
                if cls.CHECK_MANIFEST_RULES:
                    raise InvalidManifest(msg)
                else:
                    warnings.warn(msg)

            try:
                rules.append(FolderRule(folder, **folder_rule if folder_rule else {}))
            except InvalidIfClause as e:
                raise InvalidManifest(f'Invalid manifest file {path}: {e}')

        return Manifest(rules)

    def _most_suitable_rule(self, _folder: str) -> FolderRule:
        folder = os.path.abspath(_folder)
        for rule in self.rules[::-1]:
            if os.path.commonpath([folder, rule.folder]) == rule.folder:
                return rule

        return DefaultRule(folder)

    def enable_build_targets(
        self, folder: str, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        return self._most_suitable_rule(folder).enable_build_targets(default_sdkconfig_target, config_name)

    def enable_test_targets(
        self, folder: str, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        return self._most_suitable_rule(folder).enable_test_targets(default_sdkconfig_target, config_name)

    def depends_components(
        self, folder: str, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        res = self._most_suitable_rule(folder).depends_components
        if isinstance(res, list):
            return res
        return res.get_value(default_sdkconfig_target or '', config_name or '')

    def depends_filepatterns(
        self, folder: str, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        res = self._most_suitable_rule(folder).depends_filepatterns
        if isinstance(res, list):
            return res
        return res.get_value(default_sdkconfig_target or '', config_name or '')



# SPDX-FileCopyrightText: 2023-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import logging
import sys
import typing as t

from .constants import (
    BuildStage,
)


class ColoredFormatter(logging.Formatter):
    grey: str = '\x1b[37;20m'
    yellow: str = '\x1b[33;20m'
    red: str = '\x1b[31;20m'
    bold_red: str = '\x1b[31;1m'

    reset: str = '\x1b[0m'

    fmt: str = '%(asctime)s %(levelname)8s %(message)s'
    app_fmt: str = f'%(asctime)s %(levelname)8s [%(build_stage){BuildStage.max_length()}s] %(message)s'

    datefmt: str = '%Y-%m-%d %H:%M:%S'

    FORMATS: t.Dict[int, str] = {
        logging.DEBUG: f'{grey}{{}}{reset}',
        logging.INFO: '{}',
        logging.WARNING: f'{yellow}{{}}{reset}',
        logging.ERROR: f'{red}{{}}{reset}',
        logging.CRITICAL: f'{bold_red}{{}}{reset}',
    }

    def __init__(self, colored: bool = True) -> None:
        self.colored = colored
        if sys.platform == 'win32':  # does not support it
            self.colored = False

        super().__init__(datefmt=self.datefmt)

    def format(self, record: logging.LogRecord) -> str:
        if getattr(record, 'build_stage', None):
            base_fmt = self.app_fmt
        else:
            base_fmt = self.fmt

        if self.colored:
            log_fmt = self.FORMATS[record.levelno].format(base_fmt)
        else:
            log_fmt = base_fmt

        if record.levelno in [logging.WARNING, logging.ERROR]:
            record.msg = '>>> ' + str(record.msg)
        elif record.levelno == logging.CRITICAL:
            record.msg = '!!! ' + str(record.msg)

        formatter = logging.Formatter(log_fmt, datefmt=self.datefmt)
        return formatter.format(record)


def setup_logging(verbose: int = 0, log_file: t.Optional[str] = None, colored: bool = True) -> None:
    """
    Setup logging stream handler

    :param verbose: 0 - WARNING, 1 - INFO, 2 - DEBUG
    :param log_file: log file path
    :param colored: colored output or not
    :return: None
    """
    if not verbose:
        level = logging.WARNING
    elif verbose == 1:
        level = logging.INFO
    else:
        level = logging.DEBUG

    package_logger = logging.getLogger(__package__)
    package_logger.setLevel(level)
    if log_file:
        handler: logging.Handler = logging.FileHandler(log_file)
    else:
        handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(ColoredFormatter(colored))

    package_logger.addHandler(handler)
    package_logger.propagate = False  # don't propagate to root logger



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['idf_build_apps',
 'idf_build_apps.junit',
 'idf_build_apps.manifest',
 'idf_build_apps.yaml']

package_data = \
{'': ['*']}

install_requires = \
['pyparsing', 'pyyaml', 'packaging', 'pydantic~=2.0', 'argcomplete>=3']

extras_require = \
{":python_version < '3.11'": ['toml'],
 "dev:python_version < '3.8'": ['typing-extensions'],
 'doc': ['sphinx',
         'sphinx-rtd-theme',
         'sphinx_copybutton',
         'myst-parser',
         'sphinxcontrib-mermaid',
         'sphinx-argparse'],
 'test': ['pytest', 'pytest-cov']}

entry_points = \
{'console_scripts': ['idf-build-apps = idf_build_apps:main.main']}

setup(name='idf-build-apps',
      version='2.4.2',
      description='Tools for building ESP-IDF related apps.',
      author=None,
      author_email='Fu Hanxi <fuhanxi@espressif.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )



# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import logging
import os.path
import typing as t
from pathlib import (
    Path,
)

from pyparsing import (
    CaselessLiteral,
    Char,
    Combine,
    Group,
    Literal,
    MatchFirst,
    OneOrMore,
    Optional,
    ParseException,
    ParseResults,
    QuotedString,
    Word,
    alphas,
    hexnums,
    nums,
)

from ..constants import (
    ALL_TARGETS,
    IDF_PATH,
)

LOGGER = logging.getLogger(__name__)

# Group for parsing literal suffix of a numbers, e.g. 100UL
_literal_symbol = Group(CaselessLiteral('L') | CaselessLiteral('U'))
_literal_suffix = OneOrMore(_literal_symbol)

# Define name
_name = Word(alphas, alphas + nums + '_')

# Define value, either a hex, int or a string
_hex_value = Combine(Literal('0x') + Word(hexnums) + Optional(_literal_suffix).suppress())('hex_value')
_str_value = QuotedString('"')('str_value')
_int_value = Word(nums)('int_value') + ~Char('.') + Optional(_literal_suffix)('literal_suffix')

# Remove optional parenthesis around values
_value = Optional('(').suppress() + MatchFirst([_hex_value, _str_value, _int_value])('value') + Optional(')').suppress()

_define_expr = '#define' + Optional(_name)('name') + Optional(_value)


def get_defines(header_path: str) -> t.List[str]:
    defines = []
    LOGGER.debug('Reading macros from %s...', header_path)
    with open(header_path) as f:
        output = f.read()

    for line in output.split('\n'):
        line = line.strip()
        if len(line):
            defines.append(line)

    return defines


def parse_define(define_line: str) -> ParseResults:
    res = _define_expr.parseString(define_line)

    return res


class SocHeader(dict):
    CAPS_HEADER_FILEPATTERN = '*_caps.h'

    def __init__(self, target: str) -> None:
        if target != 'linux':
            soc_header_dict = self._parse_soc_header(target)
        else:
            soc_header_dict = {}

        super().__init__(**soc_header_dict)

    @staticmethod
    def _get_dirs_from_candidates(candidates: t.List[str]) -> t.List[str]:
        dirs = []
        for d in candidates:
            if not os.path.isdir(d):
                LOGGER.debug('folder "%s" not found. Skipping...', os.path.abspath(d))
            else:
                dirs.append(d)

        return dirs

    @classmethod
    def _parse_soc_header(cls, target: str) -> t.Dict[str, t.Any]:
        soc_headers_dirs = cls._get_dirs_from_candidates(
            [
                # master c5 mp
                os.path.abspath(os.path.join(IDF_PATH, 'components', 'soc', target, 'mp', 'include', 'soc')),
                # other branches
                os.path.abspath(os.path.join(IDF_PATH, 'components', 'soc', target, 'include', 'soc')),
                # release/v4.2
                os.path.abspath(os.path.join(IDF_PATH, 'components', 'soc', 'soc', target, 'include', 'soc')),
            ]
        )
        esp_rom_headers_dirs = cls._get_dirs_from_candidates(
            [
                # master c5 mp
                os.path.abspath(os.path.join(IDF_PATH, 'components', 'esp_rom', target, 'mp', target)),
                # others
                os.path.abspath(os.path.join(IDF_PATH, 'components', 'esp_rom', target)),
            ]
        )

        header_files: t.List[str] = []
        for d in [*soc_headers_dirs, *esp_rom_headers_dirs]:
            LOGGER.debug('Checking dir %s', d)
            header_files += [str(p.resolve()) for p in Path(d).glob(cls.CAPS_HEADER_FILEPATTERN)]

        output_dict = {}
        for f in header_files:
            LOGGER.debug('Checking header file %s', f)
            for line in get_defines(f):
                try:
                    res = parse_define(line)
                except ParseException:
                    LOGGER.debug('Failed to parse: %s', line)
                    continue

                if 'str_value' in res:
                    output_dict[res.name] = res.str_value
                elif 'int_value' in res:
                    output_dict[res.name] = int(res.int_value)
                elif 'hex_value' in res:
                    output_dict[res.name] = int(res.hex_value, 16)

        return output_dict


SOC_HEADERS: t.Dict[str, SocHeader] = {target: SocHeader(target) for target in ALL_TARGETS}



# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import fnmatch
import functools
import glob
import logging
import os
import shutil
import subprocess
import sys
import typing as t
from copy import (
    deepcopy,
)

from packaging.version import (
    Version,
)
from pydantic import BaseModel as _BaseModel

LOGGER = logging.getLogger(__name__)


class ConfigRule:
    def __init__(self, file_name: str, config_name: str = '') -> None:
        """
        ConfigRule represents the sdkconfig file and the config name.

        For example:

            - filename='', config_name='default' - represents the default app configuration, and gives it a name
                'default'
            - filename='sdkconfig.*', config_name=None - represents the set of configurations, names match the wildcard
                value

        :param file_name: name of the sdkconfig file fragment, optionally with a single wildcard ('*' character).
            can also be empty to indicate that the default configuration of the app should be used
        :param config_name: name of the corresponding build configuration, or None if the value of wildcard is to be
            used
        """
        self.file_name = file_name
        self.config_name = config_name


def config_rules_from_str(rule_strings: t.Optional[t.List[str]]) -> t.List[ConfigRule]:
    """
    Helper function to convert strings like 'file_name=config_name' into `ConfigRule` objects

    :param rule_strings: list of rules as strings or a single rule string
    :return: list of ConfigRules
    """
    if not rule_strings:
        return []

    rules = []
    for rule_str in to_list(rule_strings):
        items = rule_str.split('=', 2)
        rules.append(ConfigRule(items[0], items[1] if len(items) == 2 else ''))
    # '' is the default config, sort this one to the front
    return sorted(rules, key=lambda x: x.file_name)


def get_parallel_start_stop(total: int, parallel_count: int, parallel_index: int) -> t.Tuple[int, int]:
    """
    Calculate the start and stop indices for a parallel task (1-based).

    :param total: total number of tasks
    :param parallel_count: number of parallel tasks to run
    :param parallel_index: index of the parallel task to run
    :return: start and stop indices, [start, stop]
    """
    if parallel_count == 1:
        return 1, total

    num_builds_per_job = (total + parallel_count - 1) // parallel_count

    _min = num_builds_per_job * (parallel_index - 1) + 1
    _max = min(num_builds_per_job * parallel_index, total)

    return _min, _max


class BuildError(RuntimeError):
    pass


class AutocompleteActivationError(SystemExit):
    pass


class InvalidCommand(SystemExit):
    def __init__(self, msg: str) -> None:
        super().__init__('Invalid Command: ' + msg.strip())


class InvalidInput(SystemExit):
    """Invalid input from user"""


class InvalidIfClause(SystemExit):
    """Invalid if clause in manifest file"""


class InvalidManifest(SystemExit):
    """Invalid manifest file"""


def rmdir(path: str, exclude_file_patterns: t.Union[t.List[str], str, None] = None) -> None:
    if not exclude_file_patterns:
        shutil.rmtree(path, ignore_errors=True)
        return

    for root, dirs, files in os.walk(path, topdown=False):
        for f in files:
            fp = os.path.join(root, f)
            remove = True
            for pattern in to_list(exclude_file_patterns):
                if pattern and fnmatch.fnmatch(f, pattern):
                    remove = False
                    break
            if remove:
                os.remove(fp)
        for d in dirs:
            try:
                os.rmdir(os.path.join(root, d))
            except OSError:
                pass


def find_first_match(pattern: str, path: str) -> t.Optional[str]:
    for root, _, files in os.walk(path):
        res = fnmatch.filter(files, pattern)
        if res:
            return os.path.join(root, res[0])
    return None


def subprocess_run(
    cmd: t.List[str],
    log_terminal: bool = True,
    log_fs: t.Union[t.IO[str], str, None] = None,
    check: bool = False,
    additional_env_dict: t.Optional[t.Dict[str, str]] = None,
    **kwargs,
) -> int:
    """
    Subprocess.run for older python versions

    :param cmd: cmd
    :param log_terminal: print to `sys.stdout` if set to `True`
    :param log_fs: write to this file stream if not `None`
    :param check: raise `BuildError` when return code is non-zero
    :param additional_env_dict: additional environment variables
    :return: return code
    """
    LOGGER.debug('==> Running %s', ' '.join(cmd))

    subprocess_env = None
    if additional_env_dict is not None:
        subprocess_env = deepcopy(os.environ)
        subprocess_env.update(additional_env_dict)

    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=subprocess_env, **kwargs)

    def _log_stdout(fs: t.Optional[t.IO[str]] = None):
        if p.stdout:
            for line in p.stdout:
                if isinstance(line, bytes):
                    line = line.decode('utf-8')

                if log_terminal:
                    sys.stdout.write(line)

                if fs:
                    fs.write(line)

    if p.stdout:
        if log_fs:
            if isinstance(log_fs, str):
                with open(log_fs, 'a') as fa:
                    _log_stdout(fa)
            else:
                _log_stdout(log_fs)

    returncode = p.wait()
    if check and returncode != 0:
        raise BuildError(f'Command {cmd} returned non-zero exit status {returncode}')

    return returncode


_T = t.TypeVar('_T')


@t.overload
def to_list(s: None) -> None: ...


@t.overload
def to_list(s: t.Iterable[_T]) -> t.List[_T]: ...


@t.overload
def to_list(s: _T) -> t.List[_T]: ...


def to_list(s):
    """
    Turn all objects to lists

    :param s: anything
    :return:
        - ``None``, if ``s`` is None
        - itself, if ``s`` is a list
        - ``list(s)``, if ``s`` is a tuple or a set
        - ``[s]``, if ``s`` is other type

    """
    if s is None:
        return s

    if isinstance(s, list):
        return s

    if isinstance(s, set) or isinstance(s, tuple):
        return list(s)

    return [s]


@t.overload
def to_set(s: None) -> None: ...


@t.overload
def to_set(s: t.Iterable[_T]) -> t.Set[_T]: ...


@t.overload
def to_set(s: _T) -> t.Set[_T]: ...


def to_set(s):
    """
    Turn all objects to sets

    :param s: anything
    :return:
        - ``None``, if ``s`` is None
        - itself, if ``s`` is a set
        - ``set(to_list(s))``, if ``s`` is other type
    """
    if s is None:
        return s

    if isinstance(s, set):
        return s

    return set(to_list(s))


def semicolon_separated_str_to_list(s: t.Optional[str]) -> t.Optional[t.List[str]]:
    """
    Split a string by semicolon and strip each part

    Args:
        s: string to split

    Returns:
        list of strings
    """
    if s is None or s.strip() == '':
        return None

    return [p.strip() for p in s.strip().split(';') if p.strip()]


def to_absolute_path(s: str, rootpath: t.Optional[str] = None) -> str:
    rp = os.path.abspath(os.path.expanduser(rootpath or '.'))

    sp = os.path.expanduser(s)
    if os.path.isabs(sp):
        return sp
    else:
        return os.path.abspath(os.path.join(rp, sp))


def to_version(s: t.Any) -> Version:
    if isinstance(s, Version):
        return s

    try:
        return Version(str(s))
    except ValueError:
        raise InvalidInput(f'Invalid version: {s}')


def files_matches_patterns(
    files: t.Union[t.List[str], str],
    patterns: t.Union[t.List[str], str],
    rootpath: t.Optional[str] = None,
) -> bool:
    # can't match an absolute pattern with a relative path
    # change all to absolute paths
    matched_paths = set()
    for pat in [to_absolute_path(p, rootpath) for p in to_list(patterns)]:
        matched_paths.update(glob.glob(str(pat), recursive=True))

    for f in [to_absolute_path(f, rootpath) for f in to_list(files)]:
        if str(f) in matched_paths:
            return True

    return False


@functools.total_ordering
class BaseModel(_BaseModel):
    """
    BaseModel that is hashable
    """

    __EQ_IGNORE_FIELDS__: t.List[str] = []

    def __lt__(self, other: t.Any) -> bool:
        if isinstance(other, self.__class__):
            for k in self.model_dump():
                if k in self.__EQ_IGNORE_FIELDS__:
                    continue

                self_attr = getattr(self, k, '') or ''
                other_attr = getattr(other, k, '') or ''

                if self_attr != other_attr:
                    return self_attr < other_attr

                continue

            return False

        return NotImplemented

    def __eq__(self, other: t.Any) -> bool:
        if isinstance(other, self.__class__):
            # we only care the public attributes
            self_model_dump = self.model_dump()
            other_model_dump = other.model_dump()

            for _field in self.__EQ_IGNORE_FIELDS__:
                self_model_dump.pop(_field, None)
                other_model_dump.pop(_field, None)

            return self_model_dump == other_model_dump

        return NotImplemented

    def __hash__(self) -> int:
        hash_list = []
        for v in self.model_dump().values():
            if isinstance(v, list):
                hash_list.append(tuple(v))
            elif isinstance(v, dict):
                hash_list.append(tuple(v.items()))
            else:
                hash_list.append(v)

        return hash((type(self), *tuple(hash_list)))
